from ....base import TestBase
from .....messaging.interfaces.dispatcher.endpoint import (DispatcherEndpoint, FAST_ADDR_TO_INTERFACE, INTERFACES,
                                                           PREFERENCE_ORDER, guess_interface)
from .....messaging.interfaces.endpoint import Endpoint, EndpointListener
from .....messaging.interfaces.udp.endpoint import UDPv4Address


class DummyEndpointListener(EndpointListener):
    """
    This class simply listens on an endpoint and stores incoming packets in a list.
    """
    def __init__(self, endpoint):
        super(DummyEndpointListener, self).__init__(endpoint)
        self.incoming = []

    def on_packet(self, packet):
        self.incoming.append(packet)


class DummyEndpoint(Endpoint):
    """
    Non-functional endpoint for manual staging and inspection.
    """

    def __init__(self):
        super(DummyEndpoint, self).__init__(prefixlen=1)

        self.opened = False
        self.sent = []
        self.bytes_up = 0
        self.bytes_down = 0

    def assert_open(self):
        assert self.opened

    def is_open(self):
        return self.opened

    def get_address(self):
        return UDPv4Address("127.0.0.1", 1337)

    def send(self, socket_address, packet):
        self.bytes_up += len(packet)
        self.sent.append((socket_address, packet))

    async def open(self):
        self.opened = True

    def close(self):
        self.opened = False

    def reset_byte_counters(self):
        self.bytes_up = 0
        self.bytes_down = 0

    def notify_listeners(self, packet):
        socket_address, data = packet
        self.bytes_down += len(data)
        super(DummyEndpoint, self).notify_listeners(packet)


FAST_ADDR_TO_INTERFACE.clear()
FAST_ADDR_TO_INTERFACE.update({tuple: "Dummy"})
INTERFACES.clear()
INTERFACES.update({"Dummy": DummyEndpoint})
PREFERENCE_ORDER.clear()
PREFERENCE_ORDER.append("Dummy")


class TestDispatcherEndpoint(TestBase):
    """
    This class contains various tests for the DispatcherEndpoint.
    """

    RANDOM_DATA = "data"

    @staticmethod
    async def _produce_dummy():
        """
        Create and open a DispatcherEndpoint, dispatching to a dummy endpoint and listener.
        """
        endpoint = DispatcherEndpoint(["Dummy"])
        child_endpoint = list(endpoint.interfaces.values())[0]
        listener = DummyEndpointListener(endpoint)
        endpoint.add_listener(listener)
        await endpoint.open()
        return endpoint, child_endpoint, listener

    async def test_initialize_no_interfaces(self):
        """
        Check if the DispatcherEndpoint can initialize and "send" without interfaces.

        This is black-hole functionality and should not crash (though we can't really assert anything here).
        """
        endpoint = DispatcherEndpoint([])
        endpoint.send(("0.0.0.0", 0), TestDispatcherEndpoint.RANDOM_DATA)

    async def test_dispatch_receive(self):
        """
        Check if packet reception is correctly propagated from children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()
        packet = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA

        child_endpoint.notify_listeners(packet)

        self.assertEqual(1, len(listener.incoming))
        self.assertEqual(packet, listener.incoming[0])

    async def test_dispatch_send(self):
        """
        Check if packet sending is correctly propagated to children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()
        packet = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA

        endpoint.send(*packet)

        self.assertEqual(1, len(child_endpoint.sent))
        self.assertEqual(packet, child_endpoint.sent[0])

    async def test_dispatch_send_specific(self):
        """
        Check if packet sending is correctly propagated to children, with specific interface.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()
        packet = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA

        endpoint.send(*packet, interface="Dummy")

        self.assertEqual(1, len(child_endpoint.sent))
        self.assertEqual(packet, child_endpoint.sent[0])

    async def test_is_open(self):
        """
        Check if is_open is correctly propagated from children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()

        # The Child Endpoint is open and the Dispatcher Endpoint propagates the child's status.
        self.assertTrue(child_endpoint.is_open())
        self.assertTrue(endpoint.is_open())

        # Close the Dispatcher Endpoint.
        endpoint.close()

        # The Child Endpoint is closed and the Dispatcher Endpoint propagates the child's status.
        self.assertFalse(child_endpoint.is_open())
        self.assertFalse(endpoint.is_open())
        self.assertRaises(AssertionError, endpoint.assert_open)

    async def test_remove_listener(self):
        """
        Check if remove_listener is correctly propagated to children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()
        packet = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA

        # Registered listeners receive data.
        endpoint.notify_listeners(packet)
        self.assertEqual(1, len(listener.incoming))

        # Remove the listener from the Dispatcher Endpoint.
        endpoint.remove_listener(listener)

        # Unregistered listeners no longer receive data.
        endpoint.notify_listeners(packet)
        self.assertEqual(1, len(listener.incoming))

    async def test_byte_counters(self):
        """
        Check if byte counters are correctly propagated from children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()
        packet = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA

        self.assertEqual(0, endpoint.bytes_up)
        self.assertEqual(0, endpoint.bytes_down)

        # 2x RANDOM_DATA up and 1x RANDOM_DATA down.
        endpoint.send(*packet)
        endpoint.send(*packet)
        child_endpoint.notify_listeners(packet)

        self.assertEqual(2 * len(TestDispatcherEndpoint.RANDOM_DATA), endpoint.bytes_up)
        self.assertEqual(len(TestDispatcherEndpoint.RANDOM_DATA), endpoint.bytes_down)

        # Back to 0.
        endpoint.reset_byte_counters()

        self.assertEqual(0, endpoint.bytes_up)
        self.assertEqual(0, endpoint.bytes_down)

    async def test_get_address(self):
        """
        Check if get_address is correctly propagated from children.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()

        self.assertEqual(UDPv4Address("127.0.0.1", 1337), endpoint.get_address())

    async def test_get_address_specific(self):
        """
        Check if get_address is correctly propagated from children, with specific interface.
        """
        endpoint, child_endpoint, listener = await self._produce_dummy()

        self.assertEqual(UDPv4Address("127.0.0.1", 1337), endpoint.get_address(interface="Dummy"))

    async def test_add_prefix_listener(self):
        """
        Check if add_prefix_listener is correctly propagated to children.
        """
        endpoint, child_endpoint, listener1 = await self._produce_dummy()
        listener2 = DummyEndpointListener(endpoint)
        endpoint.add_prefix_listener(listener2, "s")
        packet1 = ("1.2.3.4", 5), TestDispatcherEndpoint.RANDOM_DATA
        packet2 = ("1.2.3.4", 5), "sata"

        child_endpoint.notify_listeners(packet1)
        child_endpoint.notify_listeners(packet2)

        # Prefix listener should've ignored "data" and only received "sata".
        self.assertEqual(2, len(listener1.incoming))
        self.assertEqual(1, len(listener2.incoming))

    async def test_guess_interface_ipv4(self):
        """
        Check if guess_interface guesses IPv4 interfaces correctly.
        """
        self.assertEqual("UDPIPv4", guess_interface(("1.2.3.4", 5)))

    async def test_guess_interface_ipv6(self):
        """
        Check if guess_interface guesses IPv6 interfaces correctly.
        """
        self.assertEqual("UDPIPv6", guess_interface(("2001:0db8:0000:0000:0000:ff00:0042:8329", 5)))

    async def test_guess_interface_ipv6_short(self):
        """
        Check if guess_interface guesses shortened IPv6 interfaces correctly.
        """
        self.assertEqual("UDPIPv6", guess_interface(("2001:db8:0:0:0:ff00:42:8329", 5)))

    async def test_guess_interface_ipv6_very_short(self):
        """
        Check if guess_interface guesses zero-omitted IPv6 interfaces correctly.
        """
        self.assertEqual("UDPIPv6", guess_interface(("2001:db8::ff00:42:8329", 5)))

    async def test_guess_interface_ipv6_ipv4map(self):
        """
        Check if guess_interface guesses IPv4 mapped onto IPv6 interfaces correctly.
        """
        self.assertEqual("UDPIPv6", guess_interface(("::ffff:192.0.2.128", 5)))

    async def test_guess_interface_unknown(self):
        """
        Check if guess_interface guesses None if the interface is invalid.
        """
        self.assertIsNone(guess_interface(("2001:db8:::ff00:42:8329", 5)))



import logging
from struct import unpack

from ...requestcache import NumberCache


class HashCache(NumberCache):
    """
    Cache tied to a hash.
    """

    def __init__(self, request_cache, prefix, cache_hash, id_format):
        prefix, number = self.id_from_hash(prefix, cache_hash)
        super(HashCache, self).__init__(request_cache, prefix, number)
        self.id_format = id_format

    @classmethod
    def id_from_hash(cls, prefix, cache_hash):
        number = 0
        for i in range(len(cache_hash)):
            b, = unpack('>B', cache_hash[i:i + 1])
            number <<= 8
            number |= b
        return prefix, number


class PeerCache(NumberCache):
    """
    Cache tied to a peer (mid).
    """

    def __init__(self, request_cache, prefix, mid, id_format):
        prefix, number = self.id_from_address(prefix, mid)
        super(PeerCache, self).__init__(request_cache, prefix, number)
        self.id_format = id_format

    @classmethod
    def id_from_address(cls, prefix, mid):
        return HashCache.id_from_hash(prefix, mid)

    @property
    def timeout_delay(self):
        return 120.0


class ReceiveAttestationVerifyCache(HashCache):
    """
    Pending attestation transfer, after request for attestation verification.
    """

    def __init__(self, community, cache_hash, id_format):
        super(ReceiveAttestationVerifyCache, self).__init__(community.request_cache, u"receive-verify-attestation",
                                                            cache_hash, id_format)
        self.attestation_map = set()

    def on_timeout(self):
        logging.warning("ReceiveAttestationVerify timed out!")

    @property
    def timeout_delay(self):
        return 120.0


class ReceiveAttestationRequestCache(PeerCache):
    """
    Pending attestation transfer, after request for a new attestation.
    Stores one-time key for this attribute attestation.
    """

    def __init__(self, community, mid, key, name, id_format):
        super(ReceiveAttestationRequestCache, self).__init__(community.request_cache, u"receive-request-attestation",
                                                             mid, id_format)
        self.attestation_map = set()
        self.key = key
        self.name = name

    def on_timeout(self):
        logging.warning("ReceiveAttestation timed out!")


class ProvingAttestationCache(HashCache):
    """
    Pending attestation verification, stores expected relmap, hashed challenges and completion callback.
    """

    def __init__(self, community, cache_hash, id_format, public_key=None, on_complete=lambda x, y: None):
        super(ProvingAttestationCache, self).__init__(community.request_cache, u"proving-attestation", cache_hash,
                                                      id_format)
        self.hash = cache_hash
        self.public_key = public_key
        self.relativity_map = {}
        self.hashed_challenges = []
        self.challenges = []
        self.attestation_callbacks = on_complete

    def on_timeout(self):
        logging.warning("ProvingAttestation timed out!")

    @property
    def timeout_delay(self):
        return 120.0


class PendingChallengeCache(HashCache):
    """
    Single pending challenge for a ProvingAttestationCache.
    """
    def __init__(self, community, cache_hash, proving_cache, id_format, honesty_check=-1):
        super(PendingChallengeCache, self).__init__(community.request_cache, u"proving-hash", cache_hash, id_format)
        self.proving_cache = proving_cache
        self.honesty_check = honesty_check

    def on_timeout(self):
        logging.warning("PendingChallenge timed out!")



from ...peerdiscovery.discovery import DiscoveryStrategy


class MockWalk(DiscoveryStrategy):

    def take_step(self):
        for peer in self.overlay.network.verified_peers:
            self.overlay.walk_to(peer.address)



import json
import os
from base64 import decodebytes, encodebytes
from binascii import unhexlify
from functools import wraps
from hashlib import sha1
from random import choice
from threading import RLock

from .caches import (HashCache, PeerCache, PendingChallengeCache, ProvingAttestationCache,
                     ReceiveAttestationRequestCache, ReceiveAttestationVerifyCache)
from .database import AttestationsDB
from .payload import (AttestationChunkPayload, ChallengePayload, ChallengeResponsePayload, RequestAttestationPayload,
                      VerifyAttestationRequestPayload)
from ..schema.manager import SchemaManager
from ...community import Community
from ...lazy_community import lazy_wrapper
from ...messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from ...requestcache import RequestCache
from ...util import maybe_coroutine


def synchronized(f):
    """
    Due to database inconsistencies, we can't allow multiple threads to handle a received_half_block at the same time.
    """
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        with self.receive_block_lock:
            return f(self, *args, **kwargs)
    return wrapper


class AttestationCommunity(Community):
    """
    Community for sharing Attestations.

    Note that the logic for giving out Attestations is in the identity chain.
    """
    community_id = unhexlify('b42c93d167a0fc4a0843f917d4bf1e9ebb340ec4')

    def __init__(self, *args, **kwargs):
        working_directory = kwargs.pop('working_directory', '')
        db_name = kwargs.pop('db_name', 'attestations')

        super(AttestationCommunity, self).__init__(*args, **kwargs)

        self.receive_block_lock = RLock()

        self.schema_manager = SchemaManager()
        self.schema_manager.register_default_schemas()

        self.attestation_request_callback = lambda peer, attribute_name, metadata: None
        self.attestation_request_complete_callback = \
            lambda for_peer, attribute_name, attr_hash, id_format, from_peer=None: None
        self.verify_request_callback = lambda attribute_name, attr_hash: True

        # Map of attestation hash -> (PrivateKey, id_format)
        self.attestation_keys = {}
        self.database = AttestationsDB(working_directory, db_name)
        for attribute_hash, _, key, id_format in self.database.get_all():
            attribute_hash = attribute_hash if isinstance(attribute_hash, bytes) else str(attribute_hash)
            key = key if isinstance(key, bytes) else str(key)
            id_format = (id_format if isinstance(id_format, bytes) else str(id_format)).decode('utf-8')
            self.attestation_keys[attribute_hash] = (self.get_id_algorithm(id_format).load_secret_key(key), id_format)
        self.cached_attestation_blobs = {}
        self.allowed_attestations = {}  # mid -> global_time

        self.request_cache = RequestCache()

        self.add_message_handler(VerifyAttestationRequestPayload, self.on_verify_attestation_request)
        self.add_message_handler(AttestationChunkPayload, self.on_attestation_chunk)
        self.add_message_handler(ChallengePayload, self.on_challenge)
        self.add_message_handler(ChallengeResponsePayload, self.on_challenge_response)
        self.add_message_handler(RequestAttestationPayload, self.on_request_attestation)

    async def unload(self):
        await self.request_cache.shutdown()

        await super(AttestationCommunity, self).unload()
        # Close the database after we stop accepting requests.
        self.database.close()

    def get_id_algorithm(self, id_format):
        return self.schema_manager.get_algorithm_instance(id_format)

    def set_attestation_request_callback(self, f):
        """
        Set the callback to be called when someone requests an attestation from us.

        f should accept a (Peer, attribute name, metadata) and return a str()-able value.
        If it f returns None, no attestation is made.

        :param f: the callback function providing the value
        """
        self.attestation_request_callback = f

    def set_attestation_request_complete_callback(self, f):
        """
        f should accept a (Peer, attribute_name, hash, id_format, Peer=None), it is called when an Attestation
        has been made for another peer

        :param f: the function to call when an Attestation has been completed
        """
        self.attestation_request_complete_callback = f

    def set_verify_request_callback(self, f):
        """
        Set the callback to be called when someone wants to verify our attribute.

        f should accept a (Peer, attribute_name) and return a boolean value.
        If f return True, the attribute will be verified.

        :param f: the function to call when our attribute is requested for verification
        """
        self.verify_request_callback = f

    def dump_blob(self, attribute_name, id_format, blob, metadata={}):
        """
        Add an attribute directly (without the help of an IPv8 peer).

        This is only for advanced use, where the blob already has (1) some form of attestation embedded and (2)
        follows some form of non-interactive Zero-Knowledge Proof.

        :param attribute_name: the attribute we are creating
        :param id_format: the identity format
        :param blob: the raw data to be processed by the given id_format
        :param metadata: optional additional metadata
        """
        id_algorithm = self.get_id_algorithm(id_format)
        attestation_blob, key = id_algorithm.import_blob(blob)
        attestation = id_algorithm.get_attestation_class().unserialize_private(key, attestation_blob, id_format)

        self.on_attestation_complete(attestation, key, self.my_peer, attribute_name, attestation.get_hash(), id_format)

    def request_attestation(self, peer, attribute_name, secret_key, metadata={}):
        """
        Request attestation of one of our attributes.

        :param peer: Peer of the Attestor
        :param attribute_name: the attribute we want attested
        :param secret_key: the secret key we use for this attribute
        """
        public_key = secret_key.public_key()
        id_format = metadata.pop("id_format", "id_metadata")

        meta_dict = {
            "attribute": attribute_name,
            "public_key": encodebytes(public_key.serialize()).decode(),
            "id_format": id_format
        }
        meta_dict.update(metadata)
        metadata = json.dumps(meta_dict).encode()

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        payload = RequestAttestationPayload(metadata)
        dist = GlobalTimeDistributionPayload(global_time)

        gtime_str = str(global_time).encode('utf-8')
        self.request_cache.add(ReceiveAttestationRequestCache(self, peer.mid + gtime_str, secret_key, attribute_name,
                                                              id_format))
        self.allowed_attestations[peer.mid] = (self.allowed_attestations.get(peer.mid, [])
                                               + [gtime_str])

        packet = self._ez_pack(self._prefix, 5, [auth, dist, payload])
        self.endpoint.send(peer.address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, RequestAttestationPayload)
    async def on_request_attestation(self, peer, dist, payload):
        """
        Someone wants us to attest their attribute.
        """
        metadata = json.loads(payload.metadata)
        attribute = metadata.pop('attribute')
        pubkey_b64 = metadata.pop('public_key').encode()
        id_format = metadata.pop('id_format')
        id_algorithm = self.get_id_algorithm(id_format)

        value = await maybe_coroutine(self.attestation_request_callback, peer, attribute, metadata)
        if value is None:
            return

        PK = id_algorithm.load_public_key(decodebytes(pubkey_b64))
        attestation_blob = id_algorithm.attest(PK, value)
        attestation = id_algorithm.get_attestation_class().unserialize(attestation_blob, id_format)

        self.attestation_request_complete_callback(peer, attribute, attestation.get_hash(), id_format)

        self.send_attestation(peer.address, attestation_blob, dist.global_time)

    def on_attestation_complete(self, unserialized, secret_key, peer, name, attestation_hash, id_format):
        """
        We got an Attestation delivered to us.
        """
        self.attestation_keys[attestation_hash] = (secret_key, id_format)
        self.database.insert_attestation(unserialized, attestation_hash, secret_key, id_format)
        self.attestation_request_complete_callback(self.my_peer, name, attestation_hash, id_format, peer)

    def verify_attestation_values(self, socket_address, attestation_hash, values, callback, id_format):
        """
        Ask the peer behind a socket address to deliver the Attestation with a certain hash.

        :param socket_address: the socket address to send to
        :param attestation_hash: the hash of the Attestation to request
        :param values: the values for which we want to measure certainty
        :param callback: the callback to call with the map of (hash, {value: certainty})
        :param id_format: the identity format specifier
        """
        algorithm = self.get_id_algorithm(id_format)

        def on_complete(attestation_hash, relativity_map):
            callback(attestation_hash, [algorithm.certainty(value, relativity_map) for value in values])
        self.request_cache.add(ProvingAttestationCache(self, attestation_hash, id_format, on_complete=on_complete))
        self.create_verify_attestation_request(socket_address, attestation_hash, id_format)

    def create_verify_attestation_request(self, socket_address, attestation_hash, id_format):
        """
        Ask the peer behind a socket address to deliver the Attestation with a certain hash.

        :param socket_address: the socket address to send to
        :param attestation_hash: the hash of the Attestation to request
        :param id_format: the identity format specifier
        """
        self.request_cache.add(ReceiveAttestationVerifyCache(self, attestation_hash, id_format))

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        payload = VerifyAttestationRequestPayload(attestation_hash)
        dist = GlobalTimeDistributionPayload(global_time)

        packet = self._ez_pack(self._prefix, 1, [auth, dist, payload])
        self.endpoint.send(socket_address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, VerifyAttestationRequestPayload)
    async def on_verify_attestation_request(self, peer, dist, payload):
        """
        We received a request to verify one of our attestations. Send the requested attestation back.
        """
        attestation_blob = self.database.get_attestation_by_hash(payload.hash)
        if not attestation_blob:
            self.logger.warning("Dropping verification request of unknown hash!")
            return
        attestation_blob, = attestation_blob
        if not attestation_blob:
            self.logger.warning("Attestation blob for verification is empty!")
            return

        value = await maybe_coroutine(self.verify_request_callback, peer, payload.hash)
        if not value:
            return

        SK, id_format = self.attestation_keys[payload.hash]
        attestation_cls = self.get_id_algorithm(id_format).get_attestation_class()
        private_attestation = attestation_cls.unserialize_private(SK, attestation_blob, id_format)
        public_attestation_blob = private_attestation.serialize()
        self.cached_attestation_blobs[payload.hash] = private_attestation
        self.send_attestation(peer.address, public_attestation_blob)

    def send_attestation(self, socket_address, blob, global_time=None):
        # If we want to serve this request send the attestation in chunks of 800 bytes
        sequence_number = 0
        for i in range(0, len(blob), 800):
            blob_chunk = blob[i:i + 800]
            self.logger.debug("Sending attestation chunk %d to %s", sequence_number, str(socket_address))
            if global_time is None:
                global_time = self.claim_global_time()
            auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
            payload = AttestationChunkPayload(sha1(blob).digest(), sequence_number, blob_chunk)
            dist = GlobalTimeDistributionPayload(global_time)
            packet = self._ez_pack(self._prefix, 2, [auth, dist, payload])
            self.endpoint.send(socket_address, packet)

            sequence_number += 1

    @lazy_wrapper(GlobalTimeDistributionPayload, AttestationChunkPayload)
    def on_attestation_chunk(self, peer, dist, payload):
        """
        We received a chunk of an Attestation.
        """
        hash_id = HashCache.id_from_hash(u"receive-verify-attestation", payload.hash)
        peer_ids = [PeerCache.id_from_address(u"receive-request-attestation", peer.mid + allowed_glob)
                    for allowed_glob in self.allowed_attestations.get(peer.mid, [])
                    if allowed_glob == str(dist.global_time).encode('utf-8')]
        if self.request_cache.has(*hash_id):
            cache = self.request_cache.get(*hash_id)
            cache.attestation_map |= {(payload.sequence_number, payload.data), }

            serialized = b""
            for (_, chunk) in sorted(cache.attestation_map, key=lambda item: item[0]):
                serialized += chunk

            attestation_class = self.get_id_algorithm(cache.id_format).get_attestation_class()
            if sha1(serialized).digest() == payload.hash:
                unserialized = attestation_class.unserialize(serialized, cache.id_format)
                self.request_cache.pop(*hash_id)
                self.on_received_attestation(peer, unserialized, payload.hash)

            self.logger.debug("Received attestation chunk %d for proving by %s", payload.sequence_number, str(peer))
        else:
            handled = False
            for peer_id in peer_ids:
                if self.request_cache.has(*peer_id):
                    cache = self.request_cache.get(*peer_id)
                    cache.attestation_map |= {(payload.sequence_number, payload.data), }

                    serialized = b""
                    for (_, chunk) in sorted(cache.attestation_map, key=lambda item: item[0]):
                        serialized += chunk

                    attestation_class = self.get_id_algorithm(cache.id_format).get_attestation_class()
                    if sha1(serialized).digest() == payload.hash:
                        unserialized = attestation_class.unserialize_private(cache.key, serialized, cache.id_format)
                        cache = self.request_cache.pop(*peer_id)
                        self.allowed_attestations[peer.mid] = [glob_time for glob_time
                                                               in self.allowed_attestations[peer.mid]
                                                               if glob_time != str(dist.global_time).encode('utf-8')]
                        if not self.allowed_attestations[peer.mid]:
                            self.allowed_attestations.pop(peer.mid)
                        self.on_attestation_complete(unserialized, cache.key, peer, cache.name, unserialized.get_hash(),
                                                     cache.id_format)

                    self.logger.debug("Received attestation chunk %d for my attribute %s",
                                      payload.sequence_number,
                                      cache.name)
                    handled = True
                    break
            if not handled:
                self.logger.warning("Received Attestation chunk which we did not request!")

    def on_received_attestation(self, peer, attestation, attestation_hash):
        """
        Callback for when we got the entire attestation from a peer.

        :param peer: the Peer we got this attestation from
        :param attestation: the Attestation object we can check
        """
        algorithm = self.get_id_algorithm(attestation.id_format)

        relativity_map = algorithm.create_certainty_aggregate(attestation)
        hashed_challenges = []
        cache = self.request_cache.get(*HashCache.id_from_hash(u"proving-attestation", attestation_hash))
        cache.public_key = attestation.PK
        challenges = algorithm.create_challenges(attestation.PK, attestation)
        for challenge in challenges:
            challenge_hash = sha1(challenge).digest()
            hashed_challenges.append(challenge_hash)
        cache.relativity_map = relativity_map
        cache.hashed_challenges = hashed_challenges
        cache.challenges = challenges
        self.logger.debug("Sending %d challenges to %s", len(challenges), str(peer))
        remaining = 10
        for challenge in challenges:
            if remaining == 0:
                break
            elif self.request_cache.has(*PendingChallengeCache.id_from_hash(u"proving-hash", sha1(challenge).digest())):
                continue
            remaining -= 1
            self.request_cache.add(PendingChallengeCache(self, sha1(challenge).digest(), cache, cache.id_format))

            global_time = self.claim_global_time()
            auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
            payload = ChallengePayload(attestation_hash, challenge)
            dist = GlobalTimeDistributionPayload(global_time)

            packet = self._ez_pack(self._prefix, 3, [auth, dist, payload])
            self.endpoint.send(peer.address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, ChallengePayload)
    def on_challenge(self, peer, dist, payload):
        """
        We received a challenge for an Attestation.
        """
        SK, id_format = self.attestation_keys[payload.attestation_hash]
        challenge_hash = sha1(payload.challenge).digest()
        algorithm = self.get_id_algorithm(id_format)
        attestation = self.cached_attestation_blobs[payload.attestation_hash]

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        payload = ChallengeResponsePayload(challenge_hash,
                                           algorithm.create_challenge_response(SK, attestation, payload.challenge))
        dist = GlobalTimeDistributionPayload(global_time)

        packet = self._ez_pack(self._prefix, 4, [auth, dist, payload])
        self.endpoint.send(peer.address, packet)

    @synchronized
    @lazy_wrapper(GlobalTimeDistributionPayload, ChallengeResponsePayload)
    def on_challenge_response(self, peer, dist, payload):
        """
        We received a response to our challenge
        """
        cache = self.request_cache.get(*HashCache.id_from_hash(u"proving-hash", payload.challenge_hash))
        if cache:
            self.request_cache.pop(*HashCache.id_from_hash(u"proving-hash", payload.challenge_hash))
            proving_cache = cache.proving_cache
            pcache_prefix, pcache_id = HashCache.id_from_hash(u"proving-attestation", proving_cache.hash)
            challenge = None
            if payload.challenge_hash in proving_cache.hashed_challenges:
                proving_cache.hashed_challenges.remove(payload.challenge_hash)
                for challenge in proving_cache.challenges[:]:
                    if sha1(challenge).digest() == payload.challenge_hash:
                        proving_cache.challenges.remove(challenge)
                        break
            algorithm = self.get_id_algorithm(proving_cache.id_format)
            if cache.honesty_check < 0:
                algorithm.process_challenge_response(proving_cache.relativity_map, challenge, payload.response)
            elif not algorithm.process_honesty_challenge(cache.honesty_check, payload.response):
                self.logger.error("%s tried to cheat in the ZKP!", peer.address[0])
                # Liar, Completed
                if self.request_cache.has(pcache_prefix, pcache_id):
                    self.request_cache.pop(pcache_prefix, pcache_id)
                proving_cache.attestation_callbacks(proving_cache.hash, algorithm.create_certainty_aggregate(None))
            if len(proving_cache.hashed_challenges) == 0:
                self.logger.info("Completed attestation verification")
                # Completed
                if self.request_cache.has(pcache_prefix, pcache_id):
                    self.request_cache.pop(pcache_prefix, pcache_id)
                proving_cache.attestation_callbacks(proving_cache.hash, proving_cache.relativity_map)
            else:
                # Send another proving hash
                honesty_check = algorithm.honesty_check and (os.urandom(1)[0] < 38)
                honesty_check_byte = choice([0, 1, 2]) if honesty_check else -1
                challenge = None
                if honesty_check:
                    while not challenge or self.request_cache.has(*HashCache.id_from_hash(u"proving-hash",
                                                                                          sha1(challenge).digest())):
                        challenge = algorithm.create_honesty_challenge(proving_cache.public_key, honesty_check_byte)
                if (not honesty_check) or (challenge and self.request_cache.has(*HashCache.id_from_hash(u"proving-hash",
                                                                                sha1(challenge).digest()))):
                    honesty_check_byte = -1
                    challenge = None
                    for c in proving_cache.challenges:
                        if not self.request_cache.has(*HashCache.id_from_hash(u"proving-hash", sha1(c).digest())):
                            challenge = c
                            break
                    if not challenge:
                        self.logger.debug("No more bitpairs to challenge!")
                        return
                self.logger.debug("Sending challenge: %d (%d)", honesty_check_byte,
                                  len(proving_cache.hashed_challenges))
                self.request_cache.add(PendingChallengeCache(self, sha1(challenge).digest(), proving_cache,
                                                             cache.id_format, honesty_check_byte))

                global_time = self.claim_global_time()
                auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
                payload = ChallengePayload(proving_cache.hash, challenge)
                dist = GlobalTimeDistributionPayload(global_time)

                packet = self._ez_pack(self._prefix, 3, [auth, dist, payload])
                self.endpoint.send(peer.address, packet)



from ..base import TestBase
from ..mocking.community import MockCommunity
from ...peerdiscovery.discovery import EdgeWalk


class TestEdgeWalk(TestBase):

    def setUp(self):
        super(TestEdgeWalk, self).setUp()

        node_count = 3
        self.overlays = [MockCommunity() for _ in range(node_count)]
        self.strategies = [EdgeWalk(self.overlays[i], neighborhood_size=1) for i in range(node_count)]

        # Prevent LAN IPs from ending up in Network
        for overlay in self.overlays:
            overlay.address_is_lan = lambda _: False

    async def tearDown(self):
        for overlay in self.overlays:
            await overlay.unload()
        return await super(TestEdgeWalk, self).tearDown()

    async def test_take_step(self):
        """
        Check if we will walk to a random other node.

        Unit test network layout:
          NODE0 <-> NODE1 <-> NODE2
        """
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
        self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
        # We expect NODE1 to introduce NODE0 to NODE2
        self.strategies[0].take_step()  # First it's added in the neighborhood
        self.strategies[0].take_step()  # Second it introduces its neighbor

        await self.deliver_messages()

        self.strategies[0].take_step()  # Find out the neighbor has been introduced and walk to it
        await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 2)

    async def test_take_step_into(self):
        """
        Check if we will walk to an introduced node.
        """
        self.strategies[0].edge_timeout = 0.0  # Finish the edge immediately
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])

        # We expect NODE0 will add NODE1 to its neighborhood and start constructing an edge from it.
        # Don't allow that right now.
        self.strategies[0].take_step()

        await self.deliver_messages()

        # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
        self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])

        # In order:
        # 1. Add root (NODE1) and query for nodes
        # 2. Detect intro (NODE2) from root and query for nodes
        # 3. Detect no more intros from NODE2 and finish edge
        for _ in range(3):
            self.strategies[0].take_step()  # Attempt intro
            await self.deliver_messages()
            self.strategies[0].take_step()  # Complete intro
            await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 2)
        self.assertEqual(len(self.strategies[0].complete_edges), 1)

    async def test_fail_step_into(self):
        """
        Check if we drop an unreachable introduced node.
        """
        self.strategies[0].edge_timeout = 0.0  # Finish the edge immediately
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[2].endpoint.close()

        # We expect NODE0 will add NODE1 to its neighborhood and start constructing an edge from it.
        # Don't allow that right now.
        self.strategies[0].take_step()

        await self.deliver_messages()

        # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
        self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])

        # In order:
        # 1. Add root (NODE1) and query for nodes
        # 2. Detect intro (NODE2) from root and query for nodes
        # 3. Fail to walk to NODE2 -> edge is only root, so no complete edge
        for _ in range(3):
            self.strategies[0].take_step()
            await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 1)
        self.assertEqual(len(self.strategies[0].complete_edges), 0)

    async def test_complete_edge(self):
        """
        Check if we can complete an edge.
        """
        self.strategies[0].edge_length = 2  # Finish with one other node
        self.strategies[0].edge_timeout = 1.0  # Finish the edge, but allow the network to walk
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])

        # We expect NODE0 will add NODE1 to its neighborhood and start constructing an edge from it.
        # Don't allow that right now.
        self.strategies[0].take_step()

        await self.deliver_messages()

        # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
        self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])

        # In order:
        # 1. Add root (NODE1) and query for nodes
        # 2. Detect intro (NODE2) from root and query for nodes
        # 3. Detect no more intros from NODE2 and finish edge
        for _ in range(3):
            self.strategies[0].take_step()
            await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 2)
        self.assertEqual(len(self.strategies[0].complete_edges), 1)



import asyncio
import collections
import logging
import socket

from ..endpoint import Endpoint, EndpointClosedException


UDPv4Address = collections.namedtuple("UDPv4Address", ["ip", "port"])
UDPv4LANAddress = collections.namedtuple("UDPv4LANAddress", ["ip", "port"])
UDPv6Address = collections.namedtuple("UDPv6Address", ["ip", "port"])
DomainAddress = collections.namedtuple("DomainAddress", ["host", "port"])


class UDPEndpoint(Endpoint, asyncio.DatagramProtocol):

    SOCKET_FAMILY = socket.AF_INET

    def __init__(self, port=0, ip="0.0.0.0"):
        Endpoint.__init__(self)
        # Endpoint info
        self._port = port
        self._ip = ip
        self._running = False

        # The transport object passed on by Asyncio
        self._transport = None

        # Byte counters
        self.bytes_up = 0
        self.bytes_down = 0

    def datagram_received(self, datagram, addr):
        # If the endpoint is still running, accept incoming requests, otherwise drop them
        if self._running:
            self.bytes_down += len(datagram)
            self.notify_listeners((UDPv4Address(*addr), datagram))

    def send(self, socket_address, packet):
        """
        Send a packet to a given address.
        :param socket_address: Tuple of (IP, port) which indicates the destination of the packet.
        """
        self.assert_open()
        try:
            self._transport.sendto(packet, socket_address)
            self.bytes_up += len(packet)
        except (TypeError, ValueError, AttributeError) as exc:
            self._logger.warning("Dropping packet due to message formatting error: %s", exc)

    def log_error(self, message, level=logging.WARNING):
        self._logger.log(level, message)

    async def open(self):
        """
        Open the the Endpoint.

        :return: True is the Endpoint was successfully opened, False otherwise.
        """
        # If the endpoint is already running, then there is no need to try and open it again

        if self._running:
            return True

        loop = asyncio.get_event_loop()

        for _ in range(10000):
            try:
                # It is recommended that this endpoint is opened at port = 0,
                # such that the OS handles the port assignment
                self._transport = socket.socket(self.SOCKET_FAMILY, socket.SOCK_DGRAM)
                self._transport.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 870400)
                self._transport.bind((self._ip, self._port))
                self._transport.setblocking(False)
                self._port = self._transport.getsockname()[1]

                self._transport, _ = await loop.create_datagram_endpoint(lambda: self,
                                                                         sock=self._transport)

                self._logger.debug("Listening at %d", self._port)
                break
            except (OSError, ValueError):
                self._logger.debug("Listening failed at %d", self._port)
                self._port += 1
                continue

        self._running = True
        return True

    def assert_open(self):
        if not self._running and (not self._transport or not self._transport.is_closing()):
            raise EndpointClosedException(self)

    def close(self):
        """
        Closes the Endpoint.
        """
        if not self._running:
            return

        self._running = False

        if not self._transport.is_closing():
            self._transport.close()

    def get_address(self):
        """
        Get the address for this Endpoint.
        """
        self.assert_open()
        return self._transport.get_extra_info("socket").getsockname()

    def is_open(self):
        """
        Check if the underlying socket is open.
        """
        return self._running

    def reset_byte_counters(self):
        """
        Set bytes_up and bytes_down to 0.
        """
        self.bytes_up = 0
        self.bytes_down = 0


class UDPv6Endpoint(UDPEndpoint):

    SOCKET_FAMILY = socket.AF_INET6

    def __init__(self, port=0, ip="::"):
        super(UDPv6Endpoint, self).__init__(port, ip)

    def datagram_received(self, datagram, addr):
        # If the endpoint is still running, accept incoming requests, otherwise drop them
        if self._running:
            self.bytes_down += len(datagram)
            self.notify_listeners((UDPv6Address(*addr[:2]), datagram))



from __future__ import annotations

import inspect
import types
import typing

from .serialization import FormatListType, Payload


class VariablePayload(Payload):
    """
    A Payload instance which mimics a struct. Useful for when you want a less verbose way to specify Payloads.

    This class requires you to only specify your format in:

        - <MyPayload>.format_list : a list of Serializer format specifications
        - <MyPayload>.names : the field names to use for the given formats

    For instance:

        class MyPayload(VariablePayload):
            format_list = ['?']
            names = ["is_this_a_boolean"]

    If you require field-specific pack/unpack operations you can specify them using the `fix_pack_*` and
    `fix_unpack_*` methods.
    Custom packing and unpacking rules can be useful for compression methods like socket.inet_aton, which you only
    want to apply when actually sending over the wire.
    """

    names: typing.List[str] = []

    def __init__(self, *args, **kwargs):
        """
        Instantiate this VariablePayload class.

        :param args: the anonymous list of arguments, an index-based mapping to self.names and self.format_list
        :param kwargs: the named arguments, mapping to self.names and self.format_list (in no particular order)
        :raises KeyError: if the given arguments do not match the class specification
        """
        index = 0
        fwd_args = {}
        # If our super class is an old-style Payload function, forward the required arguments.
        if not issubclass(super(VariablePayload, self).__class__, VariablePayload) and \
                inspect.ismethod(super(VariablePayload, self).__init__):
            super_argspec = inspect.getfullargspec(super(VariablePayload, self).__init__).args[1:]
            for arg in super_argspec:
                if arg in kwargs:
                    fwd_args[arg] = kwargs.pop(arg)
                else:
                    fwd_args[arg] = args[index]
                index += 1
            super(VariablePayload, self).__init__(**fwd_args)
        Payload.__init__(self)
        # Try to fill the required format specification.
        base = index
        for i in range(len(self.format_list) - index):
            # Run out all anonymous arguments, then start popping from the keyword arguments.
            # This will raise a KeyError if we provide more arguments than we can handle.
            for _ in range(8 if self.format_list[i + base] == 'bits' else 1):
                value = args[index] if index < len(args) else kwargs.pop(self.names[index])
                setattr(self, self.names[index], value)
                index += 1
        if len(args) - index > 0:
            raise KeyError("%s missing %d arguments!" % (self.__class__.__name__, len(args) - index))
        if kwargs:
            raise KeyError("%s has leftover keyword arguments: %s!" % (self.__class__.__name__, str(kwargs)))

    @classmethod
    def from_unpack_list(cls, *args) -> VariablePayload:
        """
        Given a list of raw arguments, initialize a new cls instance.

        If this class has special rules for certain fields, apply them.
        """
        unpack_args = list(args)
        for i in range(len(args)):
            custom_rule = "fix_unpack_" + cls.names[i]
            if hasattr(cls, custom_rule):
                unpack_args[i] = getattr(cls, custom_rule)(args[i])
        return cls(*unpack_args)

    @staticmethod
    def _to_packlist_fmt(fmt: FormatListType) -> str:
        if isinstance(fmt, str):
            return fmt
        if isinstance(fmt, list):
            return 'payload-list'
        return 'payload'

    def _fix_pack(self, name: str) -> typing.Any:
        """
        Check if there are custom rules for sending this field.
        """
        raw_value = getattr(self, name)
        custom_rule = "fix_pack_" + name
        if hasattr(self, custom_rule):
            return getattr(self, custom_rule)(raw_value)
        return raw_value

    def to_pack_list(self) -> typing.List[tuple]:
        """
        Convert the VariablePayload to a Serializable pack list.
        This method will automatically pull from the available format names and set instance fields.

        :return: the pack list
        """
        out = []
        index = 0
        for i in range(len(self.format_list)):
            args = []
            for _ in range(8 if self.format_list[i] == 'bits' else 1):
                args.append(self._fix_pack(self.names[index]))
                index += 1
            out.append((self._to_packlist_fmt(self.format_list[i]), *args))
        return out


def _compile_init(names: typing.List[str]) -> types.CodeType:
    """
    Compile the init function.

    Takes the form of:

     .. code-block :: Python

        def __init__(self, a, b):
            self.a = a
            self.b = b

    :param names: the format list's names
    :type names: [str]
    :return: the compiled code object
    :rtype: code
    """
    f_code = """
def __init__(self, %s):
    Payload.__init__(self)
    %s
    """ % (', '.join(names),
           '\n    '.join(['self.%s = %s' % (name, name) for name in names]))
    return compile(f_code, f_code, 'exec')


def _compile_from_unpack_list(src_cls: typing.Type[VariablePayload], names: typing.List[str]) -> types.CodeType:
    """
    Compile the unpacking code.

    Takes the form of (``fix_unpack_`` is inserted if defined in the source class):

    .. code-block :: Python

        def from_unpack_list(cls, a, b):
            return cls(a, fix_unpack_b(b), c)

    :param src_cls: the source class to use
    :param src_cls: VariablePayload
    :param names: the format list's names
    :type names: [str]
    :return: the compiled code object
    :rtype: code
    """
    arg_list = ', '.join(names)
    f_code = """
def from_unpack_list(cls, %s):
    return cls(%s)
    """ % (arg_list, ', '.join(["None if %s is None else cls.fix_unpack_%s(%s)" % (name, name, name)
                                if hasattr(src_cls, "fix_unpack_" + name)
                                else name for name in names]))
    return compile(f_code, f_code, 'exec')


def _compile_to_pack_list(src_cls: typing.Type[VariablePayload],
                          format_list: typing.List[FormatListType],
                          names: typing.List[str]) -> types.CodeType:
    """
    Compile the packing code.

    Takes the form of (``fix_pack_`` is inserted if defined in the source class):

    .. code-block :: Python

        def to_pack_list(self):
            return [("I", self.a), ("H", fix_pack_b(self.b)))]

    :param src_cls: the source class to use
    :param src_cls: VariablePayload
    :param format_list: the format_list
    :type format_list: [FormatListType]
    :param names: the format list's names
    :type names: [str]
    :return: the compiled code object
    :rtype: code
    """
    f_code = """
def to_pack_list(self):
    return [%s]
        """
    fmts = []
    index = 0
    for fmt in format_list:
        args = []
        for _ in range(8 if fmt == 'bits' else 1):
            name = names[index]
            if hasattr(src_cls, "fix_pack_" + name):
                args.append(f"self.fix_pack_{name}(self.{name})")
            else:
                args.append(f"self.{name}")
            index += 1
        derived_fmt = fmt if isinstance(fmt, str) else ("payload-list" if isinstance(fmt, list) else "payload")
        fmts.append('("%s", %s)' % (derived_fmt, ", ".join(args)))
    f_code %= ', '.join(fmts)
    return compile(f_code, f_code, 'exec')


def vp_compile(vp_definition: typing.Type[VariablePayload]) -> typing.Type[VariablePayload]:
    """
    JIT Compilation of a VariablePayload definition.
    """

    # We use ``exec`` purposefully here, disable the pylint warning:
    # pylint: disable=W0122

    # Load the function definitions into the local scope.
    exec(_compile_init(vp_definition.names), globals(), locals())
    exec(_compile_from_unpack_list(vp_definition, vp_definition.names), globals(), locals())
    exec(_compile_to_pack_list(vp_definition, vp_definition.format_list, vp_definition.names), globals(), locals())

    # Rewrite the class methods from the locally loaded overwrites.
    # from_unpack_list is a classmethod, so we need to scope it properly.
    setattr(vp_definition, '__init__', locals()['__init__'])
    setattr(vp_definition, 'from_unpack_list', types.MethodType(locals()['from_unpack_list'], vp_definition))
    setattr(vp_definition, 'to_pack_list', locals()['to_pack_list'])
    return vp_definition


__all__ = ['VariablePayload', 'vp_compile']



import functools
from threading import RLock

from aiohttp import ClientSession

from ..base import TestBase
from ...REST.rest_manager import RESTManager
from ...configuration import get_default_configuration
from ...keyvault.crypto import ECCrypto
from ...messaging.anonymization.endpoint import TunnelEndpoint
from ...peer import Peer
from ...peerdiscovery.network import Network
from ...test.mocking.discovery import MockWalk
from ...test.mocking.endpoint import AutoMockEndpoint
from ...util import maybe_coroutine


def partial_cls(cls, *args, **kwargs):
    class PartialCls(cls):
        __init__ = functools.partialmethod(cls.__init__, *args, **kwargs)
    PartialCls.__name__ = cls.__name__
    return PartialCls


class MockRestIPv8(object):
    def __init__(self, crypto_curve, overlay_classes, *args, **kwargs):
        self.endpoint = AutoMockEndpoint()
        self.endpoint.open()
        self.configuration = get_default_configuration()
        self.configuration["working_directory"] = ":memory:"
        self.network = Network()
        self.my_peer = Peer(ECCrypto().generate_key(crypto_curve))
        self.overlays = [overlay_cls(self.my_peer, self.endpoint, self.network, *args, **kwargs)
                         for overlay_cls in overlay_classes]
        self.strategies = [(MockWalk(overlay), 20) for overlay in self.overlays]
        self.rest_manager = None
        self.rest_port = 0
        self.overlay_lock = RLock()

    def get_overlay(self, overlay_cls):
        return next((o for o in self.overlays if isinstance(o, overlay_cls)), None)

    def add_strategy(self, overlay, strategy, target_peers):
        with self.overlay_lock:
            self.overlays.append(overlay)
            self.strategies.append((strategy, target_peers))

    def unload_overlay(self, instance):
        self.overlays = [overlay for overlay in self.overlays if overlay != instance]
        self.strategies = [(strategy, target_peers) for (strategy, target_peers) in self.strategies
                           if strategy.overlay != instance]
        return maybe_coroutine(instance.unload)

    async def produce_anonymized_endpoint(self):
        endpoint = TunnelEndpoint(AutoMockEndpoint())
        endpoint.open()
        return endpoint

    async def start_api(self):
        self.rest_manager = RESTManager(self)
        await self.rest_manager.start(0)
        self.rest_port = self.rest_manager.site._server.sockets[0].getsockname()[1]

    async def stop(self, stop_loop=True):
        await self.rest_manager.stop()
        self.endpoint.close()
        for overlay in self.overlays:
            await overlay.unload()


class RESTTestBase(TestBase):
    """
    HTTP request superclass, which defines the common behavior between the different types of HTTP REST requests
    """

    async def initialize(self, overlay_classes, node_count, *args, **kwargs):
        self.overlay_classes = overlay_classes
        self.nodes = [await self.create_node(*args, **kwargs) for _ in range(node_count)]

        for node in self.nodes:
            for other in self.nodes:
                if other == node:
                    continue
                private_peer = other.my_peer
                public_peer = Peer(private_peer.public_key, private_peer.address)
                for overlay_class in overlay_classes:
                    node.network.discover_services(public_peer, overlay_class.community_id)

    async def create_node(self, *args, **kwargs):
        ipv8 = MockRestIPv8(u"curve25519", overlay_classes=self.overlay_classes, *args, **kwargs)
        await ipv8.start_api()
        return ipv8

    async def introduce_nodes(self):
        for node in self.nodes:
            for other in self.nodes:
                if other != node:
                    for overlay in node.overlays:
                        overlay.walk_to(other.endpoint.wan_address)
        await self.deliver_messages()

    async def make_request(self, node, endpoint, request_type, arguments=None, json_response=True, json=None,
                           expected_status=200):
        """
        Forward an HTTP request of the specified type to a url, with the specified set of arguments.

        :param node: the destination node
        :param endpoint: the endpoint of this request (i.e. http://<interface>:<port>/<endpoint>)
        :param request_type: the type of request (GET, POST, PUT, DELETE, etc.)
        :param arguments: the arguments to be attached to the request. This should be a dictionary or None
        :param json_response: whether the response is expected to be JSON
        :param json: a JSON-serializable dictionary that is sent when making the request
        :param expected_status: the status code returned in the response, defaults to 200
        :return: a dictionary object with the response
        """

        # Using localhost in the URL will cause aiohttp to first try ::1, causing a 1s delay for each request
        url = 'http://127.0.0.1:%d/%s' % (node.rest_port, endpoint)
        headers = {'User-Agent': 'aiohttp'}

        async with ClientSession() as session:
            async with session.request(request_type, url, json=json, params=arguments, headers=headers) as response:
                self.assertEqual(response.status, expected_status,
                                 "Expected HTTP status code %d, got %d" % (expected_status, response.status))
                return await response.json(content_type=None) if json_response else await response.read()



import ipaddress
import logging
import typing

from ..endpoint import Endpoint
from ..udp.endpoint import UDPEndpoint, UDPv4Address, UDPv6Address, UDPv6Endpoint

INTERFACES = {
    "UDPIPv4": UDPEndpoint,
    "UDPIPv6": UDPv6Endpoint
}
"""
The INTERFACES dictionary describes the mapping of interface names to interface classes.
"""

PREFERENCE_ORDER = [
    "UDPIPv4",
    "UDPIPv6"
]
"""
The PREFERENCE_ORDER list describes the order of preference for the available interfaces.
For example, ``["UDPIPv4", "UDPIPv6"]`` means: use IPv4 over IPv6, if it is available.
"""


FAST_ADDR_TO_INTERFACE = {
    UDPv4Address: "UDPIPv4",
    UDPv6Address: "UDPIPv6"
}
"""
The FAST_ADDR_TO_INTERFACE is an internal DispatcherEndpoint mapping, to quickly route information.
For addresses which do not use these classes the slower ``guess_interface`` will be used.
"""


def guess_interface(socket_address: typing.Any):
    """
    Attempt to guess the interface for the given address.

    If the given address is a tuple of a valid IPv4 address string and a port this returns "UDPIPv4".
    If the given address is a tuple of a valid IPv6 address string and a port this returns "UDPIPv6".
    Otherwise, this returns None.
    """
    try:
        if (isinstance(socket_address, tuple) and len(socket_address) == 2
                and isinstance(socket_address[0], str) and isinstance(socket_address[1], int)):
            if isinstance(ipaddress.ip_address(socket_address[0]), ipaddress.IPv4Address):
                return "UDPIPv4"
            else:
                return "UDPIPv6"
    except Exception:
        logging.exception("Exception occurred while guessing interface for %s", repr(socket_address))
    return None


class DispatcherEndpoint(Endpoint):
    """
    An Endpoint implementation to dispatch to other Endpoint implementations.

    The added complexity is as follows:

     - Receiving of packets is hooked directly into the sub-Endpoints, no speed is lost.
     - Sending packets will be directed to the appropriate interface. If an address object class is not defined in
       FAST_ADDR_TO_INTERFACE, this will have to use ``guess_interface``.
     - Adding and removing listeners will have to be forwarded to all sub-Endpoints.
    """

    def __init__(self, interfaces: typing.List[str], **kwargs) -> None:
        """
        Create a new DispatcherEndpoint by giving a list of interfaces to load (check INTERFACES for available
        interfaces).

        You can optionally supply keyword arguments to launch each selected Endpoint implementation, for example:

         .. code-block :: Python

            DispatcherEndpoint(["UDPIPv4"], UDPIPv4={'port': my_custom_port})

        :param interfaces: list of interfaces to load.
        :param kwargs: optional interface-specific launch arguments.
        :returns: None
        """
        super(DispatcherEndpoint, self).__init__()
        # Filter the available interfaces and preference order, based on the user's selection.
        self.interfaces = {interface: INTERFACES[interface](**(kwargs.get(interface, {}))) for interface in interfaces}
        self.interface_order = [interface for interface in PREFERENCE_ORDER if interface in interfaces]
        # The order of preference will not change, we can precompute the preferred interface Endpoint.
        self._preferred_interface = self.interfaces[self.interface_order[0]] if self.interface_order else None

    @property
    def bytes_up(self):
        return sum(interface.bytes_up for interface in self.interfaces.values())

    @property
    def bytes_down(self):
        return sum(interface.bytes_down for interface in self.interfaces.values())

    def add_listener(self, listener) -> None:
        for interface in self.interfaces.values():
            interface.add_listener(listener)

    def add_prefix_listener(self, listener, prefix) -> None:
        for interface in self.interfaces.values():
            interface.add_prefix_listener(listener, prefix)

    def remove_listener(self, listener) -> None:
        for interface in self.interfaces.values():
            interface.remove_listener(listener)

    def notify_listeners(self, packet) -> None:
        for interface in self.interfaces.values():
            interface.notify_listeners(packet)

    def assert_open(self) -> None:
        assert self.is_open()

    def is_open(self) -> bool:
        return any(interface.is_open() for interface in self.interfaces.values())

    def get_address(self, interface=None) -> typing.Optional[typing.Any]:
        if interface is not None:
            return self.interfaces[interface].get_address()
        elif self._preferred_interface:
            return self._preferred_interface.get_address()
        return None

    def send(self, socket_address, packet, interface=None) -> None:
        if interface is not None:
            self.interfaces[interface].send(socket_address, packet)
        else:
            interface = self.interfaces.get(FAST_ADDR_TO_INTERFACE.get(socket_address.__class__)
                                            or guess_interface(socket_address))
            if interface is not None:
                interface.send(socket_address, packet)

    async def open(self) -> None:
        for interface in self.interfaces.values():
            await interface.open()

    def close(self) -> None:
        for interface in self.interfaces.values():
            interface.close()

    def reset_byte_counters(self) -> None:
        for interface in self.interfaces.values():
            interface.reset_byte_counters()



import time

from .base import TestDHTBase
from ..mocking.ipv8 import MockIPv8
from ...dht import DHTError
from ...dht.discovery import DHTDiscoveryCommunity
from ...dht.routing import Node, RoutingTable
from ...util import succeed


class TestDHTDiscoveryCommunity(TestDHTBase):

    def setUp(self):
        super().setUp()
        self.initialize(DHTDiscoveryCommunity, 2)
        self.pinged = None
        self.puncture_to = None

        for node in self.nodes:
            node.overlay.cancel_pending_task('store_peer')
            node.overlay.token_maintenance()

    def create_node(self, *args, **kwargs):
        return MockIPv8(u"curve25519", DHTDiscoveryCommunity)

    async def test_store_peer(self):
        await self.introduce_nodes()
        await self.overlay(0).store_peer()
        self.assertIn(self.mid(0), self.overlay(1).store)
        self.assertIn(self.mid(0), self.overlay(0).store_for_me)

    async def test_store_peer_fail(self):
        await self.introduce_nodes()
        self.overlay(0).routing_tables[self.address(0).__class__] = RoutingTable(self.my_node_id(0))
        self.assertFalse(await self.overlay(0).store_peer())

    async def test_connect_peer(self):
        # Add a third node
        node = MockIPv8(u"curve25519", DHTDiscoveryCommunity)
        self.add_node_to_experiment(node)
        await self.introduce_nodes()

        # Node1 is storing the peer of node0
        self.overlay(1).store[self.mid(0)].append(self.my_peer(0))
        self.overlay(0).store_for_me[self.mid(0)].append(self.my_peer(1))

        org_func = self.overlay(1).create_puncture_request

        def create_puncture_request(*args):
            self.puncture_to = args[1]
            return org_func(*args)
        self.overlay(1).create_puncture_request = create_puncture_request

        await self.deliver_messages()
        nodes = await self.overlay(2).connect_peer(self.mid(0))
        self.assertEqual(self.puncture_to, self.address(2))
        self.assertIn(self.key_bin(0),
                      [n.public_key.key_to_bin() for n in nodes])

    async def test_connect_peer_fail(self):
        await self.introduce_nodes()
        self.overlay(0).routing_tables[self.address(0).__class__] = RoutingTable(self.my_node_id(0))
        with self.assertRaises(DHTError):
            await self.overlay(0).connect_peer(self.mid(1))

    async def test_ping_pong(self):
        now = time.time() - 1

        node0 = Node(self.private_key(0), self.address(0))
        node0.last_response = now
        node0.last_queries.append(now)

        node1 = Node(self.private_key(1), self.address(1))
        node1.last_response = now
        node1.last_queries.append(now)

        key = node1.mid
        self.overlay(0).store[key].append(node1)
        self.overlay(1).store_for_me[key].append(node0)

        await self.overlay(1).ping(node0)
        self.assertNotEqual(node0.last_response, now)
        self.assertNotEqual(node1.last_query, now)

    def test_ping_all(self):
        self.overlay(0).ping = lambda n: setattr(self, 'pinged', n) or succeed(None)

        node1 = Node(self.private_key(1), self.address(1))
        node1.last_ping_sent = time.time()
        node1.last_queries.append(time.time())

        self.overlay(0).store[node1.mid].append(node1)
        self.overlay(0).ping_all()
        self.assertIn(node1, self.overlay(0).store[node1.mid])

        node1.last_queries[-1] -= 100
        self.overlay(0).ping_all()
        self.assertNotIn(node1, self.overlay(0).store[node1.mid])
        self.assertEqual(self.pinged, None)

        self.overlay(0).store_for_me[node1.mid].append(node1)
        self.overlay(0).ping_all()
        self.assertIn(node1.mid, self.overlay(0).store_for_me)

        node1.last_ping_sent -= 30
        self.overlay(0).ping_all()
        self.assertEqual(self.pinged, node1)
        self.assertIn(node1, self.overlay(0).store_for_me[node1.mid])

        self.pinged = None
        node1.failed = 3
        self.overlay(0).ping_all()
        self.assertEqual(self.pinged, None)
        self.assertNotIn(node1, self.overlay(0).store_for_me[node1.mid])



from binascii import unhexlify

from ...base import TestBase
from ....messaging.anonymization.tunnel import DataChecker

tracker_pkt = unhexlify('00000417271019800000000012345678')
dht_pkt = b'd1:ad2:id20:abcdefghij01234567899:info_hash20:mnopqrstuvwxyz123456e1:q9:get_peers1:t2:aa1:y1:qe'
utp_pkt = unhexlify('210086446ed69ec1ddbd9e6000100000f32e86be')
ipv8_pkt = unhexlify('0002123456789abcdef123456789abcdef123456789a00000001')
tunnel_pkt = unhexlify('000281ded07332bdc775aa5a46f96de9f8f390bbc9f300000001')


class TestDataChecker(TestBase):

    def test_could_be_dht(self):
        """
        Check if a DHT packet is correctly identified.
        """
        self.assertFalse(DataChecker.could_be_dht(tracker_pkt))
        self.assertTrue(DataChecker.could_be_dht(dht_pkt))
        self.assertFalse(DataChecker.could_be_dht(utp_pkt))
        self.assertFalse(DataChecker.could_be_dht(ipv8_pkt))
        self.assertFalse(DataChecker.could_be_dht(tunnel_pkt))

    def test_could_be_udp_tracker(self):
        """
        Check if a UDP tracker packet is correctly identified.
        """
        self.assertTrue(DataChecker.could_be_udp_tracker(tracker_pkt))
        self.assertFalse(DataChecker.could_be_udp_tracker(dht_pkt))
        self.assertFalse(DataChecker.could_be_udp_tracker(utp_pkt))
        self.assertFalse(DataChecker.could_be_udp_tracker(ipv8_pkt))
        self.assertFalse(DataChecker.could_be_udp_tracker(tunnel_pkt))

    def test_could_be_utp(self):
        """
        Check if a UTP packet is correctly identified.
        """
        self.assertFalse(DataChecker.could_be_utp(tracker_pkt))
        self.assertFalse(DataChecker.could_be_utp(dht_pkt))
        self.assertTrue(DataChecker.could_be_utp(utp_pkt))
        self.assertFalse(DataChecker.could_be_utp(ipv8_pkt))
        self.assertFalse(DataChecker.could_be_utp(tunnel_pkt))

    def test_could_be_ipv8(self):
        """
        Check if a IPv8 packet is correctly identified.
        """
        self.assertFalse(DataChecker.could_be_ipv8(tracker_pkt))
        self.assertFalse(DataChecker.could_be_ipv8(dht_pkt))
        self.assertFalse(DataChecker.could_be_ipv8(utp_pkt))
        self.assertTrue(DataChecker.could_be_ipv8(ipv8_pkt))
        self.assertTrue(DataChecker.could_be_ipv8(tunnel_pkt))

    def test_could_be_bt(self):
        """
        Check if a BitTorrent packet is correctly identified.
        """
        self.assertTrue(DataChecker.could_be_bt(tracker_pkt))
        self.assertTrue(DataChecker.could_be_bt(dht_pkt))
        self.assertTrue(DataChecker.could_be_bt(utp_pkt))
        self.assertFalse(DataChecker.could_be_bt(ipv8_pkt))
        self.assertFalse(DataChecker.could_be_bt(tunnel_pkt))



from __future__ import annotations

import binascii
import struct
import typing

from ..signed_object import AbstractSignedObject
from ...types import Metadata, PrivateKey, PublicKey


class Attestation(AbstractSignedObject):
    """
    A pointer to Metadata.

    An Attestation does not and should not contain an index.
    An Attestation does not and should not contain a reference to the public key (directly).
    """

    def __init__(self,
                 metadata_pointer: bytes,
                 private_key: typing.Optional[PrivateKey] = None,
                 signature: typing.Optional[bytes] = None):
        self.metadata_pointer = metadata_pointer
        super(Attestation, self).__init__(private_key, signature)

    def get_plaintext(self) -> bytes:
        return self.metadata_pointer

    @classmethod
    def unserialize(cls, data: bytes, public_key: PublicKey, offset: int = 0) -> Attestation:
        sig_len = public_key.get_signature_length()
        metadata_pointer, signature = struct.unpack_from(f">32s{sig_len}s", data, offset=offset)
        return Attestation(metadata_pointer, signature=signature)

    @classmethod
    def create(cls, metadata: Metadata, private_key: PrivateKey) -> Attestation:
        return Attestation(metadata.get_hash(), private_key=private_key)

    def to_database_tuple(self) -> typing.Tuple[bytes, bytes]:
        """
        Get a representation of this Attestation as two byte strings (metadata hash and signature).

        :returns: the two byte strings for database insertion.
        """
        return self.metadata_pointer, self.signature

    @classmethod
    def from_database_tuple(cls,
                            metadata_pointer: bytes,
                            signature: bytes) -> Attestation:
        """
        Create a Token from a two-byte-string representation (metadata hash and signature).

        :param metadata_pointer: the hash of the Attestation.
        :param signature: the signature over the plaintext Attestation.
        """
        return Attestation(metadata_pointer, signature=signature)

    def __str__(self) -> str:
        return f"Attestation({binascii.hexlify(self.metadata_pointer).decode()})"



from ....base import TestBase
from .....attestation.wallet.primitives.ec import esum, weilpairing
from .....attestation.wallet.primitives.value import FP2Value


class TestPairing(TestBase):

    def test_small_weilpairing(self):
        """
        Check if Weil pairing in E[4] mod 11 of (5, 4) and (5x, 4) with S=(7, 6) equals 9 + 7x
        """
        mod = 11
        wp = weilpairing(mod,
                         4,
                         (FP2Value(mod, 5), FP2Value(mod, 4)),
                         (FP2Value(mod, b=5), FP2Value(mod, 4)),
                         (FP2Value(mod, 7), FP2Value(mod, 6)))

        self.assertEqual(wp.a, 9)
        self.assertEqual(wp.b, 7)
        self.assertEqual(wp.c, 0)
        self.assertEqual(wp.aC, 1)
        self.assertEqual(wp.bC, 0)
        self.assertEqual(wp.cC, 0)

    def test_medium_weilpairing(self):
        """
        Check if Weil pairing in E[408] mod 1223 of (764, 140) and (18x, 84) with S=(0, 1222) equals 438 + 50x
        """
        mod = 1223
        wp = weilpairing(mod,
                         408,
                         (FP2Value(mod, 764), FP2Value(mod, 140)),
                         (FP2Value(mod, b=18), FP2Value(mod, 84)),
                         (FP2Value(mod, 0), FP2Value(mod, 1222)))

        self.assertEqual(wp.a, 438)
        self.assertEqual(wp.b, 50)
        self.assertEqual(wp.c, 0)
        self.assertEqual(wp.aC, 1)
        self.assertEqual(wp.bC, 0)
        self.assertEqual(wp.cC, 0)

    def test_oob_esum(self):
        """
        Check if EC sum of the point of infinity with itself is the point at infinity.
        """
        self.assertEqual(esum(11, "O", "O"), "O")

    def test_spob_esum(self):
        """
        Check if EC sum of the point of infinity with another point equals the other point.
        """
        p = (FP2Value(11, 1), FP2Value(11, 2))

        self.assertEqual(esum(11, p, "O"), p)
        self.assertEqual(esum(11, "O", p), p)



import logging
from asyncio import BaseTransport, DatagramProtocol, get_event_loop
from asyncio.futures import Future
from binascii import hexlify
from socket import AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_BROADCAST, SO_REUSEADDR, socket
from typing import Coroutine, Iterable, Optional, Union

from ..bootstrapper_interface import Bootstrapper
from ...types import Address, Community
from ...util import succeed


PROTOCOL_VERSION = b'\x00\x00'
MAGIC = b'\x49\x50\x76\x38'

HDR_ANNOUNCE = PROTOCOL_VERSION + MAGIC + b'\x00'


class BroadcastBootstrapEndpoint(DatagramProtocol):

    def __init__(self, overlay: Community):
        super().__init__()

        self._socket: Optional[socket] = None
        self._transport: Optional[BaseTransport] = None
        self.overlay = overlay
        self.logger = logging.getLogger(self.__class__.__name__)

    async def open(self) -> bool:
        loop = get_event_loop()

        try:
            self._socket = socket(AF_INET, SOCK_DGRAM)
            self._socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
            self._socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
            self._socket.bind(('', 0))
            self._transport, _ = await loop.create_datagram_endpoint(lambda: self, sock=self._socket)
        except (OSError, ValueError):
            return False

        return True

    def send(self, socket_address, data: bytes):
        try:
            if self._socket is not None:
                self._socket.sendto(data, socket_address)
        except (TypeError, ValueError, AttributeError, OSError):
            pass  # Windows doesn't really care, Ubuntu throws an exception on "illegal" ports

    def datagram_received(self, data: bytes, addr) -> None:
        if data.startswith(HDR_ANNOUNCE):
            if self.overlay.get_prefix() == data[len(HDR_ANNOUNCE):]:
                self.logger.debug("Received data from beacon %s: attempting walk!", repr(addr))
                self.overlay.walk_to(addr)
            # Otherwise: valid, but not for our overlay
        elif data.startswith(self.overlay.get_prefix()):
            self.logger.debug("Walk success by %s: attempting handoff!", repr(addr))
            self.overlay.on_packet((addr, data))
        else:
            self.logger.debug("Dropping garbage packet from %s: %s", repr(addr), hexlify(data))

    def close(self) -> None:
        if self._transport is not None and not self._transport.is_closing():
            self._transport.close()


class UDPBroadcastBootstrapper(Bootstrapper):

    def __init__(self):
        self.endpoint = None
        self.overlay = None

    async def initialize(self, overlay: Community) -> Union[Future, Coroutine]:  # pylint: disable=W0236
        self.overlay = overlay

        # Open the socket
        endpoint = BroadcastBootstrapEndpoint(overlay)
        success = await endpoint.open()
        if not success:
            return succeed(False)
        self.endpoint = endpoint

        # Start sending
        self.beacon(overlay.get_prefix())

        return succeed(True)

    def beacon(self, service_prefix: bytes) -> None:
        """
        Try to find a listener (fire and forget).
        """
        if self.endpoint is not None:
            for p in range(65535):
                self.endpoint.send(('255.255.255.255', p), HDR_ANNOUNCE + service_prefix)

    async def get_addresses(self, overlay: Community, timeout: float) -> Iterable[Address]:
        self.beacon(overlay.get_prefix())
        return []

    def keep_alive(self, overlay: Community) -> None:
        self.beacon(overlay.get_prefix())

    def blacklist(self) -> Iterable[Address]:
        return []

    def unload(self) -> None:
        if self.endpoint:
            self.endpoint.close()



import logging

from aiohttp import web

from aiohttp_apispec import AiohttpApiSpec

from .base_endpoint import HTTP_INTERNAL_SERVER_ERROR, HTTP_UNAUTHORIZED, Response
from .root_endpoint import RootEndpoint


@web.middleware
class ApiKeyMiddleware(object):
    def __init__(self, api_key):
        self.api_key = api_key

    async def __call__(self, request, handler):
        if self.authenticate(request):
            return await handler(request)
        else:
            return Response({'error': 'Unauthorized access'}, status=HTTP_UNAUTHORIZED)

    def authenticate(self, request):
        if request.path.startswith('/docs') or request.path.startswith('/static'):
            return True
        # The api key can either be in the headers or as part of the url query
        api_key = request.headers.get('X-Api-Key') or request.query.get('apikey')
        return not self.api_key or self.api_key == api_key


@web.middleware
async def cors_middleware(request, handler):
    preflight_cors = request.method == "OPTIONS" and 'Access-Control-Request-Method' in request.headers
    if not preflight_cors:
        return await handler(request)

    response = web.StreamResponse()
    # For now, just allow all methods
    response.headers['Access-Control-Allow-Methods'] = "GET, PUT, POST, PATCH, DELETE, OPTIONS"
    response.headers['Access-Control-Allow-Headers'] = '*'
    response.headers['Access-Control-Allow-Origin'] = '*'
    response.headers['Access-Control-Max-Age'] = str(86400)
    return response


@web.middleware
async def error_middleware(request, handler):
    try:
        response = await handler(request)
    except Exception as e:
        import traceback
        traceback.print_exc()
        return Response({
            "success": False,
            "error": {
                "code": e.__class__.__name__,
                "message": str(e)
            }
        }, status=HTTP_INTERNAL_SERVER_ERROR)
    return response


class RESTManager:
    """
    This class is responsible for managing the startup and closing of the HTTP API.
    """

    def __init__(self, session, root_endpoint_class=None):
        self._logger = logging.getLogger(self.__class__.__name__)
        self.session = session
        self.site = None
        self.root_endpoint = None
        self._root_endpoint_class = root_endpoint_class or RootEndpoint

    async def start(self, port=8085, host='127.0.0.1', api_key=None, ssl_context=None):
        """
        Starts the HTTP API with the listen port as specified in the session configuration.
        """
        self.root_endpoint = self._root_endpoint_class(middlewares=[ApiKeyMiddleware(api_key),
                                                                    cors_middleware,
                                                                    error_middleware])
        self.root_endpoint.initialize(self.session)

        # Not using setup_aiohttp_apispec here, as we need access to the APISpec to set the security scheme
        aiohttp_apispec = AiohttpApiSpec(
            app=self.root_endpoint.app,
            title="IPv8 REST API documentation",
            version="v2.7",  # Do not change manually! Handled by github_increment_version.py
            url="/docs/swagger.json",
            swagger_path="/docs",
        )
        if api_key:
            # Set security scheme and apply to all endpoints
            aiohttp_apispec.spec.options['security'] = [{'apiKey': []}]
            aiohttp_apispec.spec.components.security_scheme('apiKey', {'type': 'apiKey',
                                                                       'in': 'header',
                                                                       'name': 'X-Api-Key'})

        from apispec.core import VALID_METHODS_OPENAPI_V2
        if 'head' in VALID_METHODS_OPENAPI_V2:
            VALID_METHODS_OPENAPI_V2.remove('head')

        runner = web.AppRunner(self.root_endpoint.app, access_log=None)
        await runner.setup()
        # If localhost is used as hostname, it will randomly either use 127.0.0.1 or ::1
        self.site = web.TCPSite(runner, host, port, ssl_context=ssl_context)
        await self.site.start()

    async def stop(self):
        """
        Stop the HTTP API and return when the server has shut down.
        """
        await self.site.stop()



import libnacl
import libnacl.dual
import libnacl.sign

from ...keyvault.keys import PrivateKey
from ...keyvault.public.libnaclkey import LibNaCLPK


class LibNaCLSK(PrivateKey, LibNaCLPK):
    """
    A LibNaCL implementation of a secret key.
    """

    def __init__(self, binarykey=""):
        """
        Create a new LibNaCL secret key. Optionally load it from a string representation.
        Otherwise generate it from the 25519 curve.

        :param binarykey: load the sk from this string (see key_to_bin())
        """
        # Load the key, if specified
        if binarykey:
            crypt, seed = (binarykey[:libnacl.crypto_box_SECRETKEYBYTES],
                           binarykey[libnacl.crypto_box_SECRETKEYBYTES: libnacl.crypto_box_SECRETKEYBYTES
                                     + libnacl.crypto_sign_SEEDBYTES])
            self.key = libnacl.dual.DualSecret(crypt, seed)
        else:
            self.key = libnacl.dual.DualSecret()
        # Immediately create a verifier
        self.veri = libnacl.sign.Verifier(self.key.hex_vk())

    def pub(self):
        """
        Get the public key for this secret key.
        """
        return LibNaCLPK(pk=self.key.pk, hex_vk=self.veri.hex_vk())

    def signature(self, msg):
        """
        Create a signature for a message.

        :param msg: the message to sign
        :return: the signature for the message
        """
        return self.key.signature(msg)

    def key_to_bin(self):
        """
        Get the string representation of this key.
        """
        return b"LibNaCLSK:" + self.key.sk + self.key.seed



from asyncio import FIRST_COMPLETED, wait
from statistics import mean, median
from timeit import default_timer

from ipv8.messaging.anonymization.tunnel import CIRCUIT_STATE_CLOSING


async def run_speed_test(tc, circuit, request_size, response_size, num_requests, window=50):
    num_sent = 0
    num_ack = 0
    outstanding = set()
    start = default_timer()
    rtts = []

    while True:
        while num_sent < num_requests and len(outstanding) < window and circuit.state != CIRCUIT_STATE_CLOSING:
            outstanding.add(tc.send_test_request(circuit, request_size, response_size))
            num_sent += 1
        if not outstanding:
            break
        done, outstanding = await wait(outstanding, return_when=FIRST_COMPLETED, timeout=10)
        if not done:
            # We have received nothing for the past 10s.Any pending messages are considered lost.
            break
        # Make sure to only count futures that haven't been set by on_timeout.
        results = [f.result() for f in done if f.result() is not None]
        num_ack += len(results)
        rtts.extend([rtt for _, rtt in results])

    return {'speed_up': (num_ack * request_size / 1024) / (default_timer() - start),
            'speed_down': (num_ack * response_size / 1024) / (default_timer() - start),
            'messages_sent': num_ack + len(outstanding),
            'messages_received': num_ack,
            'rtt_mean': mean(rtts) if rtts else -1,
            'rtt_median': median(rtts) if rtts else -1}



from marshmallow import Schema
from marshmallow.fields import Boolean, Integer, List, Nested, String
from marshmallow.schema import SchemaMeta


class DefaultResponseSchema(Schema):
    success = Boolean(description='Indicator of success/failure', required=True)
    error = String(description='Optional field describing any failures that may have occurred', required=False)


class TransactionSchema(Schema):
    down = Integer()
    total_down = Integer()
    up = Integer()
    total_up = Integer()


class BlockSchema(Schema):
    transaction = Nested(TransactionSchema)
    linked = Nested('BlockSchema')
    type = String()
    tx = String()
    public_key = String()
    sequence_number = Integer()
    link_public_key = String()
    link_sequence_number = Integer()
    previous_hash = String()
    signature = String()
    insert_time = String()
    block_timestamp = Integer()
    block_hash = String()


class Address(Schema):
    ip = String()
    port = Integer()


class AddressWithPK(Address):
    public_key = String()


class OverlayStatisticsSchema(Schema):
    num_up = Integer()
    num_down = Integer()
    bytes_up = Integer()
    bytes_down = Integer()
    diff_time = Integer()


class OverlayStrategySchema(Schema):
    name = String()
    target_peers = Integer()


class OverlaySchema(Schema):
    id = String()
    my_peer = String()
    global_time = Integer()
    peers = List(Nested(AddressWithPK))
    overlay_name = String()
    max_peers = Integer()
    is_isolated = Boolean()
    my_estimated_wan = Nested(Address)
    my_estimated_lan = Nested(Address)
    strategies = List(Nested(OverlayStrategySchema))
    statistics = Nested(OverlayStatisticsSchema)


class DHTValueSchema(Schema):
    public_key = String()
    key = String()
    value = String()


def schema(**kwargs):
    """
    Create a schema. Mostly useful for creating single-use schemas on-the-fly.
    """
    items = list(kwargs.items())
    if len(items) != 1 or not isinstance(items[0][1], dict):
        raise RuntimeError('schema required 1 keyword argument of type dict')
    name, spec = items[0]

    schema_dict = {}
    for key, value in spec.items():
        cls, description = value if isinstance(value, tuple) else (value, None)
        required = key.endswith('*')
        key = key.rstrip('*')
        kwargs = {'required': required, 'description': description}

        if isinstance(cls, SchemaMeta):
            schema_dict[key] = Nested(cls, required=required)
        elif isinstance(cls, list) and len(cls) == 1:
            cls = cls[0]
            schema_dict[key] = List(Nested(cls), **kwargs) if isinstance(cls, SchemaMeta) else List(cls, **kwargs)
        else:
            schema_dict[key] = cls.__call__(**kwargs) if callable(cls) else cls
    return type(name, (Schema,), schema_dict)



import abc
import logging

from .keyvault.crypto import default_eccrypto
from .messaging.interfaces.endpoint import EndpointListener
from .messaging.serialization import Serializer
from .taskmanager import TaskManager


class Overlay(EndpointListener, TaskManager, metaclass=abc.ABCMeta):
    """
    Interface for an Internet overlay.
    """

    def __init__(self, community_id, my_peer, endpoint, network):
        """
        Create a new overlay for the Internet.

        :param community_id: the byte-string used to identify this overlay.
        :param my_peer: the (private key) peer of this peer
        :param endpoint: the endpoint to use for messaging
        :param network: the network graph backend
        """
        EndpointListener.__init__(self, endpoint)
        TaskManager.__init__(self)
        self.serializer = self.get_serializer()
        self.crypto = default_eccrypto

        self.community_id = community_id
        self.my_peer = my_peer

        self.endpoint.add_listener(self)

        self.logger = logging.getLogger(self.__class__.__name__)

        self.network = network

    async def unload(self):
        """
        Called when this overlay needs to shut down.
        """
        self.endpoint.remove_listener(self)
        await self.shutdown_task_manager()

    def get_serializer(self):
        """
        Get a Serializer for this Overlay.
        """
        return Serializer()

    @abc.abstractmethod
    def on_packet(self, packet):
        """
        Callback for when data is received on this endpoint.

        :param packet: the received packet, in (source, binary string) format.
        """
        pass

    @property
    def global_time(self):
        return self.my_peer.get_lamport_timestamp()

    def claim_global_time(self):
        """
        Increments the current global time by one and returns this value.
        """
        self.update_global_time(self.global_time + 1)
        return self.global_time

    def update_global_time(self, global_time):
        """
        Increase the local global time if the given GLOBAL_TIME is larger.
        """
        if global_time > self.global_time:
            self.my_peer.update_clock(global_time)

    def get_available_strategies(self):
        """
        Supply custom DiscoveryStrategies for use with this Overlay.
        This is used by the configuration system to allow for non-globally defined strategies.

        :return: a dictionary of names and DiscoveryStrategy subclass classes
        :rtype: {str: class<DiscoveryStrategy>}
        """
        return {}

    def bootstrap(self):
        """
        Perform introduction logic to get into the network.
        """
        pass

    @abc.abstractmethod
    def walk_to(self, address):
        """
        Puncture the NAT of an address.

        :param address: the address to walk to (ip, port)
        """
        pass

    @abc.abstractmethod
    def get_new_introduction(self, from_peer=None):
        """
        Get a new IP address to walk to from a random, or selected peer.

        :param from_peer: the peer to ask for an introduction
        """
        pass

    @abc.abstractmethod
    def get_peers(self):
        """
        Get the peers for this specific overlay.

        :return: the peers in the Network that use this overlay
        """
        pass

    @abc.abstractmethod
    def get_walkable_addresses(self):
        """
        Get the list of IPv4 addresses we can walk to on this overlay.

        :return: a list of IPv4 addresses
        :rtype: [(str, int)]
        """
        pass

    def get_peer_for_introduction(self, exclude=None):
        """
        Get a peer for introduction.

        :param: exclude: optionally specify a peer that is not considered eligible for introduction
        :return: a Peer to send an introduction request to, or None if there are no available
        """
        pass



import abc
import hashlib
import struct
import typing

from ..keyvault.crypto import ECCrypto
from ..types import PrivateKey, PublicKey

T = typing.TypeVar('T')


class AbstractSignedObject(metaclass=abc.ABCMeta):
    """
    To reach immutability different objects will have to be signed to reach non-repudiation.
    Examples are Tokens, Metadata and Attestation objects.

    This class handles many of the commonly required interactions for this type of signed data.
    """

    def __init__(self,
                 private_key: typing.Optional[PrivateKey] = None,
                 signature: typing.Optional[bytes] = None):
        """
        Create a new object that can be serialized and signed.
        Call this after the data has been established for the `get_plaintext()` method.
        """
        self._hash = b''
        self.crypto = ECCrypto()
        self._sign(private_key, signature)

    def get_hash(self) -> bytes:
        """
        Return the hash of this object itself.
        """
        return self._hash

    def verify(self, public_key: PublicKey) -> bool:
        """
        Verify if a public key belongs to this object.

        :returns: whether the given public key has signed for this object.
        """
        return self.crypto.is_valid_signature(public_key, self.get_plaintext(), self.signature)

    def _sign(self,
              private_key: typing.Optional[PrivateKey] = None,
              signature: typing.Optional[bytes] = None) -> None:
        """
        Add a signature to this data.
        Supply either your private key for signing or pass an existing signature.

        :param private_key: the private key to sign with.
        :param signature: the signature to adapt.
        """
        if private_key is not None and signature is None:
            self.signature = private_key.signature(self.get_plaintext())
        elif private_key is None and signature is not None:
            self.signature = signature
        else:
            raise RuntimeError("Specify either private_key or signature!")
        self._hash = hashlib.sha3_256(self.get_plaintext_signed()).digest()

    @abc.abstractmethod
    def get_plaintext(self) -> bytes:
        """
        Retrieve the content that needs to be signed, in serialized form (bytes).
        """
        pass

    def get_plaintext_signed(self) -> bytes:
        """
        Concatenate the signature to the plaintext.
        """
        return self.get_plaintext() + self.signature

    @classmethod
    @abc.abstractmethod
    def unserialize(cls: T, data: bytes, public_key: PublicKey, offset: int = 0) -> T:
        pass

    def __hash__(self) -> int:
        return struct.unpack(">Q", self._hash[:8])[0]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.get_plaintext_signed() == other.get_plaintext_signed()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return True
        return self.get_plaintext_signed() != other.get_plaintext_signed()



"""
the community module provides the Community base class that should be used when a new Community is
implemented.  It provides a simplified interface between the Dispersy instance and a running
Community instance.

@author: Boudewijn Schoon
@organization: Technical University Delft
@contact: dispersy@frayja.com
"""
import sys
from asyncio import ensure_future, iscoroutine
from binascii import hexlify
from functools import partial
from random import choice, random
from time import time
from traceback import format_exception

from .lazy_community import EZPackOverlay, lazy_wrapper, lazy_wrapper_unsigned
from .messaging.anonymization.endpoint import TunnelEndpoint
from .messaging.interfaces.dispatcher.endpoint import FAST_ADDR_TO_INTERFACE, INTERFACES
from .messaging.interfaces.udp.endpoint import UDPv4Address, UDPv4LANAddress, UDPv6Address
from .messaging.payload import (IntroductionRequestPayload, IntroductionResponsePayload,
                                NewIntroductionRequestPayload, NewIntroductionResponsePayload,
                                NewPuncturePayload, NewPunctureRequestPayload,
                                PuncturePayload, PunctureRequestPayload)
from .messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from .peer import Peer

_UNUSED_FLAGS_REQ = {'flag0': 0, 'flag1': 0, 'flag2': 0, 'flag3': 0, 'flag4': 0, 'flag5': 0, 'flag6': 0, 'flag7': 0}
_UNUSED_FLAGS_RESP = {'flag1': 0, 'flag2': 0, 'flag3': 0, 'flag4': 0, 'flag5': 0, 'flag6': 0, 'flag7': 0}

DEFAULT_MAX_PEERS = 30


class Community(EZPackOverlay):

    version = b'\x02'
    community_id = b''

    def __init__(self, my_peer, endpoint, network, max_peers=DEFAULT_MAX_PEERS, anonymize=False):
        super().__init__(self.community_id, my_peer, endpoint, network)

        self._prefix = b'\x00' + self.version + self.community_id
        self.endpoint.remove_listener(self)
        self.endpoint.add_prefix_listener(self, self._prefix)
        self.logger.debug("Launching %s with prefix %s.", self.__class__.__name__, hexlify(self._prefix).decode())

        self.max_peers = max_peers
        self.anonymize = anonymize

        if anonymize:
            if isinstance(self.endpoint, TunnelEndpoint):
                self.endpoint.set_anonymity(self._prefix, True)
            else:
                self.logger.warning('Cannot anonymize community traffic without TunnelEndpoint')

        self.network.register_service_provider(self.community_id, self)
        self.network.blacklist_mids.append(my_peer.mid)

        self.bootstrappers = []
        self._bootstrappers_initialized = False

        self.decode_map = [None] * 256

        self.add_message_handler(PunctureRequestPayload, self.on_old_puncture_request)
        self.add_message_handler(PuncturePayload, self.on_puncture)
        self.add_message_handler(NewPunctureRequestPayload, self.on_new_puncture_request)
        self.add_message_handler(NewPuncturePayload, self.on_new_puncture)
        self.add_message_handler(IntroductionRequestPayload, self.on_old_introduction_request)
        self.add_message_handler(IntroductionResponsePayload, self.on_old_introduction_response)
        self.add_message_handler(NewIntroductionRequestPayload, self.on_new_introduction_request)
        self.add_message_handler(NewIntroductionResponsePayload, self.on_new_introduction_response)

        self.add_message_handler(255, self.on_deprecated_message)
        self.add_message_handler(254, self.on_deprecated_message)
        self.add_message_handler(253, self.on_deprecated_message)
        self.add_message_handler(252, self.on_deprecated_message)
        self.add_message_handler(251, self.on_deprecated_message)
        self.add_message_handler(248, self.on_deprecated_message)
        self.add_message_handler(247, self.on_deprecated_message)
        self.add_message_handler(244, self.on_deprecated_message)
        self.add_message_handler(243, self.on_deprecated_message)
        self.add_message_handler(242, self.on_deprecated_message)
        self.add_message_handler(241, self.on_deprecated_message)
        self.add_message_handler(240, self.on_deprecated_message)
        self.add_message_handler(239, self.on_deprecated_message)
        self.add_message_handler(238, self.on_deprecated_message)
        self.add_message_handler(237, self.on_deprecated_message)
        self.add_message_handler(236, self.on_deprecated_message)
        self.add_message_handler(235, self.on_deprecated_message)

        self.deprecated_message_names = {
            255: "reserved-255",
            254: "on-missing-sequence",
            253: "missing-proof",
            252: "signature-request",
            251: "signature-response",
            248: "on-identity",
            247: "on-missing-identity",
            244: "destroy-community",
            243: "authorize",
            242: "revoke",
            241: "subjective-set",
            240: "missing-subjective-set",
            239: "on-missing-message",
            238: "undo-own",
            237: "undo-other",
            236: "dynamic-settings",
            235: "missing-last-message"
        }

    def get_prefix(self):
        return self._prefix

    async def unload(self):
        while self.bootstrappers:
            bootstrapper = self.bootstrappers.pop()
            bootstrapper.unload()
        await super().unload()

    def add_message_handler(self, msg_num, callback):
        """
        Add a handler for a message identifier. Any messages coming in with this identifier will be delivered to
        the specified callback function.

        :param msg_num: the message id to listen for (or a Payload object with a msg_id field)
        :type msg_num: int or object
        :param callback: the callback function for this message id
        :type callback: function
        :returns: None
        """
        if not isinstance(msg_num, int):
            if not hasattr(msg_num, "msg_id"):
                raise RuntimeError("Attempted to add a handler for Payload %s, which does not specify a msg_id!"
                                   % msg_num)
            msg_num = msg_num.msg_id
        if msg_num < 0 or msg_num > 255:
            raise RuntimeError("Attempted to add a handler for message number %d, which is not a byte!" % msg_num)
        if self.decode_map[msg_num]:
            raise RuntimeError("Attempted to add a handler for message number %d, already mapped to %s!" %
                               (msg_num, self.decode_map[msg_num]))
        self.decode_map[msg_num] = callback

    def on_deprecated_message(self, source_address, data):
        self.logger.warning("Received deprecated message: %s from (%s, %d)",
                            self.deprecated_message_names[data[22]], *source_address)

    def _append_bootstrapper_later(self, bootstrapper, future):
        if future.result():
            self.bootstrappers.append(bootstrapper)

    def _initialize_bootstrappers(self):
        """
        Wipes all Bootstrapper instances in self.bootstrappers.
        Adds all of the aforementioned instances back into self.bootstrappers if/when ``initialize()`` succeeded.
        """
        to_initialize = self.bootstrappers[:]
        self.bootstrappers = []
        for bootstrapper in to_initialize:
            future = ensure_future(bootstrapper.initialize(self))
            # If the bootstrapper.initialize() is not async, we would like to add the bootstrapper immediately.
            # Note that add_done_callback() will also add the bootstrapper, but in the next asyncio event (slow!).
            if future.done():
                self._append_bootstrapper_later(bootstrapper, future)
            else:
                task = self.register_anonymous_task(f"wait for bootstrap init {repr(bootstrapper)}", future)
                task.add_done_callback(partial(self._append_bootstrapper_later, bootstrapper))
        self._bootstrappers_initialized = True

    def bootstrap(self):
        if not self._bootstrappers_initialized:
            self._initialize_bootstrappers()
        for bootstrapper in self.bootstrappers:
            task = self.register_anonymous_task(f"bootstrap {repr(bootstrapper)}", bootstrapper.get_addresses,
                                                self, 60.0)
            task.add_done_callback(self.received_bootstrapped_addresses)

    def received_bootstrapped_addresses(self, addresses):
        if addresses and addresses.result():
            for address in list(addresses.result())[:self.max_peers]:
                self.walk_to(address)

    def ensure_blacklisted(self, address):
        if address not in self.network.blacklist:
            self.network.blacklist.append(address)

    def guess_address(self, interface):
        if interface == "UDPIPv4":
            return UDPv4Address(*self._get_lan_address())
        elif interface == "UDPIPv6":
            return UDPv6Address(self.get_ipv6_address(), self.endpoint.get_address(interface)[1])
        else:
            return None

    def my_preferred_address(self):
        interfaces = getattr(self.endpoint, "interfaces", [])
        if not interfaces:
            return self.my_estimated_wan
        for interface in interfaces:
            if INTERFACES[interface] not in self.my_peer.addresses:
                self.my_peer.address = self.guess_address(interface)
        return self.my_peer.address

    def create_introduction_request(self, socket_address, extra_bytes=b'', new_style=False, prefix=None):
        global_time = self.claim_global_time() % 65536
        if new_style or isinstance(socket_address, UDPv6Address):
            payload = NewIntroductionRequestPayload(socket_address, self.my_estimated_lan, self.my_preferred_address(),
                                                    global_time, **_UNUSED_FLAGS_REQ, extra_bytes=extra_bytes)
        else:
            # Only supports IPv4 addresses as tuple or UDPv4Address instances.
            payload = IntroductionRequestPayload(socket_address,
                                                 self.my_estimated_lan,
                                                 self.my_estimated_wan,
                                                 True,
                                                 u"unknown",
                                                 global_time,
                                                 extra_bytes,
                                                 supports_new_style=new_style)
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        return self._ez_pack(prefix or self._prefix, payload.msg_id, [auth, dist, payload])

    def create_introduction_response(self, lan_socket_address, socket_address, identifier,
                                     introduction=None, extra_bytes=b'', prefix=None, new_style=False):
        global_time = self.claim_global_time() % 65536
        introduction_lan = ("0.0.0.0", 0)
        introduction_wan = ("0.0.0.0", 0)
        introduced = False
        other = self.network.get_verified_by_address(socket_address)
        if not introduction:
            introduction = self.get_peer_for_introduction(exclude=other, new_style=new_style)
        if introduction:
            if isinstance(introduction.address, UDPv4Address) and self.address_is_lan(introduction.address[0]):
                introduction_lan = introduction.address
                introduction_wan = (self.my_estimated_wan[0], introduction_lan[1])
            else:
                introduction_lan = introduction.addresses.get(UDPv4LANAddress, introduction_lan)
                introduction_wan = introduction.address
            introduced = True
        new_style_intro = introduction.new_style_intro if introduction else False
        if new_style:
            payload = NewIntroductionResponsePayload(socket_address, self.my_estimated_lan, self.my_preferred_address(),
                                                     introduction_lan, introduction_wan, identifier,
                                                     new_style_intro, **_UNUSED_FLAGS_RESP, extra_bytes=extra_bytes)
        else:
            payload = IntroductionResponsePayload(socket_address,
                                                  self.my_estimated_lan,
                                                  self.my_estimated_wan,
                                                  introduction_lan,
                                                  introduction_wan,
                                                  "unknown",
                                                  identifier,
                                                  extra_bytes,
                                                  intro_supports_new_style=new_style_intro,
                                                  peer_limit_reached=0 <= self.max_peers <= len(self.get_peers()))
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        if introduced:
            packet = self.create_puncture_request(lan_socket_address, socket_address, identifier, prefix=prefix,
                                                  new_style=new_style)
            self.endpoint.send(introduction.address, packet)

        return self._ez_pack(prefix or self._prefix, payload.msg_id, [auth, dist, payload])

    def create_puncture(self, lan_walker, wan_walker, identifier, new_style=False):
        global_time = self.claim_global_time()
        if new_style:
            payload = NewPuncturePayload(lan_walker, wan_walker, identifier)
        else:
            payload = PuncturePayload(lan_walker, wan_walker, identifier)
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        return self._ez_pack(self._prefix, payload.msg_id, [auth, dist, payload])

    def create_puncture_request(self, lan_walker, wan_walker, identifier, prefix=None, new_style=False):
        global_time = self.claim_global_time()
        if new_style or not isinstance(lan_walker, UDPv4Address) or not isinstance(wan_walker, UDPv4Address):
            payload = NewPunctureRequestPayload(lan_walker, wan_walker, identifier)
        else:
            payload = PunctureRequestPayload(lan_walker, wan_walker, identifier)
        dist = GlobalTimeDistributionPayload(global_time)

        return self._ez_pack(prefix or self._prefix, payload.msg_id, [dist, payload], False)

    def introduction_request_callback(self, peer, dist, payload):
        """
        Callback that gets called after an introduction-request has been processed.
        If you want to some action to trigger upon receipt of an introduction-request,
        this would be the place to do so.

        :param peer the peer that send us an introduction-request
        :param dist the GlobalTimeDistributionPayload
        :param payload the IntroductionRequestPayload
        """
        pass

    def introduction_response_callback(self, peer, dist, payload):
        """
        Callback that gets called after an introduction-response has been processed.
        If you want to some action to trigger upon receipt of an introduction-response,
        this would be the place to do so.

        :param peer the peer that send us an introduction-response
        :param dist the GlobalTimeDistributionPayload
        :param payload the IntroductionResponsePayload
        """
        pass

    @lazy_wrapper(GlobalTimeDistributionPayload, IntroductionRequestPayload)
    def on_old_introduction_request(self, peer, dist, payload):
        if payload.supports_new_style:
            peer.new_style_intro = True
        self.on_introduction_request(peer, dist, payload)

    @lazy_wrapper(GlobalTimeDistributionPayload, NewIntroductionRequestPayload)
    def on_new_introduction_request(self, peer, dist, payload):
        peer.new_style_intro = True
        self.on_introduction_request(peer, dist, payload)

    def on_introduction_request(self, peer, dist, payload):
        if self.max_peers >= 0 and len(self.get_peers()) > self.max_peers:
            self.logger.debug("Dropping introduction request from (%s, %d): too many peers!",
                              peer.address[0], peer.address[1])
            return

        if isinstance(payload.source_lan_address, UDPv4Address):
            peer.address = UDPv4LANAddress(*payload.source_lan_address)
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [self.community_id, ])

        packet = self.create_introduction_response(payload.destination_address, peer.address, payload.identifier,
                                                   new_style=peer.new_style_intro)
        self.endpoint.send(peer.address, packet)

        self.introduction_request_callback(peer, dist, payload)

    @lazy_wrapper(GlobalTimeDistributionPayload, IntroductionResponsePayload)
    def on_old_introduction_response(self, peer, dist, payload):
        if payload.supports_new_style:
            peer.new_style_intro = True
        self.on_introduction_response(peer, dist, payload)

    @lazy_wrapper(GlobalTimeDistributionPayload, NewIntroductionResponsePayload)
    def on_new_introduction_response(self, peer, dist, payload):
        peer.new_style_intro = True
        self.on_introduction_response(peer, dist, payload)

    def on_introduction_response(self, peer, dist, payload):
        if (isinstance(payload.destination_address, UDPv4Address)
                and not self.address_in_lan_subnets(payload.destination_address[0])):
            self.my_estimated_wan = payload.destination_address
        self.my_peer.address = payload.destination_address

        if peer.new_style_intro:
            # Peer wants to use a different interface which we support, so let's try to switch interfaces.
            requested_interface = payload.source_wan_address.__class__
            used_interface = peer.address.__class__
            if (requested_interface != used_interface
                    and payload.source_wan_address != peer.address
                    and FAST_ADDR_TO_INTERFACE[requested_interface] in self.endpoint.interfaces):
                self.network.discover_address(peer, payload.source_wan_address, self.community_id, True)

                my_address = self.my_peer.addresses.get(requested_interface,
                                                        self.guess_address(FAST_ADDR_TO_INTERFACE[requested_interface]))
                if my_address:
                    packet = self.create_puncture_request(("0.0.0.0", 0), my_address, payload.identifier,
                                                          new_style=True)
                    self.endpoint.send(peer.address, packet)

        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [self.community_id, ])

        introductions = []

        if (payload.wan_introduction_address != ("0.0.0.0", 0)
                and payload.wan_introduction_address[0] != self.my_estimated_wan[0]):
            if payload.lan_introduction_address != ("0.0.0.0", 0):
                introductions.append(payload.lan_introduction_address)
            introductions.append(payload.wan_introduction_address)
        elif (payload.lan_introduction_address != ("0.0.0.0", 0)
              and payload.wan_introduction_address[0] == self.my_estimated_wan[0]):
            introductions.append(payload.lan_introduction_address)
        elif payload.wan_introduction_address != ("0.0.0.0", 0):
            introductions.append(payload.wan_introduction_address)
            introductions.append(UDPv4Address(self.my_estimated_lan[0], payload.wan_introduction_address[1]))

        for introduction in introductions:
            self.network.discover_address(peer, introduction, self.community_id, payload.intro_supports_new_style)

        self.introduction_response_callback(peer, dist, payload)

    @lazy_wrapper(GlobalTimeDistributionPayload, PuncturePayload)
    def on_puncture(self, peer, dist, payload):
        pass

    @lazy_wrapper(GlobalTimeDistributionPayload, NewPuncturePayload)
    def on_new_puncture(self, peer, dist, payload):
        pass

    @lazy_wrapper_unsigned(GlobalTimeDistributionPayload, PunctureRequestPayload)
    def on_old_puncture_request(self, source_address, dist, payload):
        self.on_puncture_request(source_address, dist, payload)

    @lazy_wrapper_unsigned(GlobalTimeDistributionPayload, NewPunctureRequestPayload)
    def on_new_puncture_request(self, source_address, dist, payload):
        self.on_puncture_request(source_address, dist, payload, True)

    def on_puncture_request(self, source_address, dist, payload, new_style=False):
        target = payload.wan_walker_address
        if payload.wan_walker_address[0] == self.my_estimated_wan[0]:
            target = payload.lan_walker_address

        packet = self.create_puncture(self.my_estimated_lan, payload.wan_walker_address, payload.identifier,
                                      new_style)
        self.endpoint.send(target, packet)

    def on_packet(self, packet, warn_unknown=True):
        source_address, data = packet
        probable_peer = self.network.get_verified_by_address(source_address)
        if probable_peer:
            probable_peer.last_response = time()
        if self._prefix != data[:22]:
            return
        msg_id = data[22]
        if self.decode_map[msg_id]:
            handler = self.decode_map[msg_id]
            try:
                result = handler(source_address, data)
                if iscoroutine(result):
                    self.register_anonymous_task('on_packet', ensure_future(result), ignore=(Exception,))
            except Exception:
                self.logger.error("Exception occurred while handling packet!\n"
                                  + ''.join(format_exception(*sys.exc_info())))
        elif warn_unknown:
            self.logger.warning("Received unknown message: %d from (%s, %d)", msg_id, *source_address)

    def walk_to(self, address):
        packet = self.create_introduction_request(address, new_style=self.network.is_new_style(address))
        self.endpoint.send(address, packet)

    def send_introduction_request(self, peer):
        """
        Send an introduction request to a specific peer.
        """
        packet = self.create_introduction_request(peer.address, new_style=peer.new_style_intro)
        self.endpoint.send(peer.address, packet)

    def get_new_introduction(self, from_peer: Peer = None):
        """
        Get a new introduction, or bootstrap if there are no available peers.
        """
        if not from_peer:
            available = self.get_peers()
            if available:
                # With a small chance, try to remedy any disconnected network phenomena.
                if self._bootstrappers_initialized and random() < 0.05:
                    for bootstrapper in self.bootstrappers:
                        bootstrapper.keep_alive(self)
                    return
                else:
                    from_peer = choice(available)
            else:
                self.bootstrap()
                return
        self.send_introduction_request(from_peer)

    def get_peer_for_introduction(self, exclude=None, new_style=False):
        """
        Return a random peer to send an introduction request to.
        """
        available = [p for p in self.get_peers() if p != exclude and not
                     (not new_style and not isinstance(p.address, UDPv4Address))]
        return choice(available) if available else None

    def get_walkable_addresses(self):
        return self.network.get_walkable_addresses(self.community_id)

    def get_peers(self):
        return self.network.get_peers_for_service(self.community_id)



"""
Copyright (c) 2016, Maarten Everts
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/gabi
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""
from random import randint

from cryptography.hazmat.primitives.asymmetric.rsa import _modinv  # type:ignore

from .credential import Credential
from .keys import CLSignature, DefaultSystemParameters, signMessageBlockAndCommitment
from .proofs import ProofS, ProofU, createChallenge, hashCommit
from .. import secure_randint
from ...primitives.value import FP2Value


class Issuer(object):

    def __init__(self, Sk, Pk, Context):
        self.Sk = Sk
        self.Pk = Pk
        self.Context = Context

    def IssueSignature(self, U, attributes, nonce2):
        signature = self.signCommitmentAndAttributes(U, attributes)
        proof = self.proveSignature(signature, nonce2)
        return IssueSignatureMessage(signature, proof)

    def signCommitmentAndAttributes(self, U, attributes):
        return signMessageBlockAndCommitment(self.Sk, self.Pk, U, [0] + attributes)

    def randomElementMultiplicativeGroup(self, modulus):
        r = 0
        while r <= 0 or _modinv(r, modulus) == 1:
            r = randint(1, modulus - 1)
        return r

    def proveSignature(self, signature, nonce2):
        Q = FP2Value(self.Pk.N, signature.A).intpow(signature.E).a
        groupModulus = self.Sk.PPrime * self.Sk.QPrime
        d = FP2Value(groupModulus, signature.E).inverse().normalize().a

        eCommit = self.randomElementMultiplicativeGroup(groupModulus)
        ACommit = FP2Value(self.Pk.N, Q).intpow(eCommit).a

        c = hashCommit([self.Context, Q, signature.A, nonce2, ACommit], False)
        eResponse = (eCommit - c * d) % groupModulus

        return ProofS(c, eResponse)


def GetProofU(pl, n):
    count = 0
    for proof in pl:
        if isinstance(proof, ProofU):
            if count == n:
                return proof
            count += 1
    return None


def GetFirstProofU(pl):
    return GetProofU(pl, 0)


def challengeContributions(pl, publicKeys, context, nonce):
    contributions = []
    for i in range(len(pl)):
        proof = pl[i]
        contributions.extend(proof.ChallengeContribution(publicKeys[i]))
    return contributions


def Verify(pl, publicKeys, context, nonce, issig, keyshareServers=[]):
    if not pl or len(pl) != len(publicKeys) or (len(keyshareServers) > 0 and len(pl) != len(keyshareServers)):
        return False

    secretkeyResponses = {}

    contributions = challengeContributions(pl, publicKeys, context, nonce)
    expectedChallenge = createChallenge(context, nonce, contributions, issig)

    kss = ""

    for i in range(len(pl)):
        proof = pl[i]
        if not proof.VerifyWithChallenge(publicKeys[i], expectedChallenge):
            return False
        if len(keyshareServers) > 0:
            kss = keyshareServers[i]
        if kss not in secretkeyResponses:
            secretkeyResponses[kss] = proof.SecretKeyResponse()
        else:
            if secretkeyResponses[kss] != proof.SecretKeyResponse():
                return False

    return True


def Challenge(builders, context, nonce, issig):
    skCommitment = secure_randint(DefaultSystemParameters[1024].LmCommit)

    commitmentValues = []
    for pb in builders:
        commitmentValues.extend(pb.Commit(skCommitment))

    return createChallenge(context, nonce, commitmentValues, issig)


def BuildDistributedProofList(builders, challenge, proofPs):
    if proofPs and len(builders) != len(proofPs):
        return None

    proofs = []

    for i in range(len(builders)):
        v = builders[i]
        proofs.append(v.CreateProof(challenge))
        if proofPs and proofPs[i]:
            proofs[i].MergeProofP(proofPs[i], v.PublicKey())

    return proofs


def BuildProofList(builders, context, nonce, issig):
    challenge = Challenge(builders, context, nonce, issig)
    return BuildDistributedProofList(builders, challenge, [])


class IssueCommitmentMessage(object):

    def __init__(self, U, Proofs, Nonce2, ProofPjwt=None, ProofPjwts=None):
        self.U = U
        self.Nonce2 = Nonce2
        self.Proofs = Proofs
        self.ProofPjwt = ProofPjwt
        self.ProofPjwts = ProofPjwts


class IssueSignatureMessage(object):

    def __init__(self, Signature, Proof):
        self.Proof = Proof
        self.Signature = Signature


def commitmentToSecret(pk, secret):
    vPrime = secure_randint(pk.Params.LvPrime)

    Sv = FP2Value(pk.N, pk.S).intpow(vPrime).a
    R0s = FP2Value(pk.N, pk.R[0]).intpow(secret).a

    return vPrime, (Sv * R0s) % pk.N


class CredentialBuilder(object):

    def __init__(self, pk, context, secret, nonce2):
        vPrime, U = commitmentToSecret(pk, secret)
        self.pk = pk
        self.context = context
        self.secret = secret
        self.vPrime = vPrime
        self.u = U
        self.uCommit = 1
        self.nonce2 = nonce2

        self.proofPcomm = None
        self.skRandomizer = None
        self.vPrimeCommit = None

    def CommitToSecretAndProve(self, nonce1):
        proofU = self.proveCommitment(self.u, nonce1)
        return IssueCommitmentMessage(self.u, [proofU], self.nonce2)

    def CreateIssueCommitmentMessage(self, proofs):
        return IssueCommitmentMessage(self.u, proofs, self.nonce2)

    def ConstructCredential(self, msg, attributes):
        if not msg.Proof.Verify(self.pk, msg.Signature, self.context, self.nonce2):
            return None

        signature = CLSignature(msg.Signature.A, msg.Signature.E, msg.Signature.V + self.vPrime)
        if self.proofPcomm:
            signature.KeyshareP = self.proofPcomm.P

        exponents = [self.secret] + attributes

        if not signature.Verify(self.pk, exponents):
            return None

        return Credential(self.pk, exponents, signature)

    def proveCommitment(self, U, nonce1):
        sCommit = secure_randint(self.pk.Params.LsCommit)
        vPrimeCommit = secure_randint(self.pk.Params.LvPrimeCommit)

        Sv = FP2Value(self.pk.N, self.pk.S).intpow(vPrimeCommit).a
        R0s = FP2Value(self.pk.N, self.pk.R[0]).intpow(sCommit).a
        Ucommit = (Sv * R0s) % self.pk.N

        c = hashCommit([self.context, U, Ucommit, nonce1], False)
        sResponse = (c * self.secret) + sCommit
        vPrimeResponse = (c * self.vPrime) + vPrimeCommit

        return ProofU(U, c, vPrimeResponse, sResponse)

    def MergeProofPCommitment(self, commitment):
        self.proofPcomm = commitment
        self.uCommit = (self.uCommit * commitment.Pcommit) % self.pk.N

    def PublicKey(self):
        return self.pk

    def Commit(self, skRandomizer):
        self.skRandomizer = skRandomizer
        self.vPrimeCommit = secure_randint(self.pk.Params.LvPrimeCommit)

        sv = FP2Value(self.pk.N, self.pk.S).intpow(self.vPrimeCommit).a
        r0s = FP2Value(self.pk.N, self.pk.R[0]).intpow(self.skRandomizer).a
        self.uCommit = (self.uCommit * sv * r0s) % self.pk.N

        ucomm = self.u
        if self.proofPcomm:
            ucomm = (ucomm * self.proofPcomm.P) % self.pk.N

        return [ucomm, self.uCommit]

    def CreateProof(self, challenge):
        sResponse = self.skRandomizer + challenge * self.secret
        vPrimeResponse = self.vPrimeCommit + challenge * self.vPrime

        return ProofU(self.u, challenge, vPrimeResponse, sResponse)



from socket import inet_aton, inet_ntoa
from struct import pack, unpack

from ..messaging.payload import IntroductionRequestPayload, Payload, decode_connection_type, encode_connection_type


class SimilarityRequestPayload(Payload):

    msg_id = 1
    format_list = ['H', '4SH', '4SH', 'bits', 'raw']

    def __init__(self, identifier, lan_address, wan_address, connection_type, preference_list):
        super(SimilarityRequestPayload, self).__init__()
        self.identifier = identifier % 65536
        self.preference_list = preference_list
        self.lan_address = lan_address
        self.wan_address = wan_address
        self.connection_type = connection_type

    def to_pack_list(self):
        encoded_connection_type = encode_connection_type(self.connection_type)
        data = [('H', self.identifier),
                ('4SH', inet_aton(self.lan_address[0]), self.lan_address[1]),
                ('4SH', inet_aton(self.wan_address[0]), self.wan_address[1]),
                ('bits', encoded_connection_type[0], encoded_connection_type[1], 0, 0, 0, 0, 0, 0),
                ('raw', b"".join(self.preference_list))]

        return data

    @classmethod
    def from_unpack_list(cls, identifier, lan_address, wan_address,
                         connection_type_0, connection_type_1, dflag0, dflag1, dflag2, dflag3, dflag4, dflag5,
                         preference_list):
        args = [identifier,
                (inet_ntoa(lan_address[0]), lan_address[1]),
                (inet_ntoa(wan_address[0]), wan_address[1]),
                decode_connection_type(connection_type_0, connection_type_1),
                [preference_list[i:i + 20] for i in range(0, len(preference_list), 20)]]

        return SimilarityRequestPayload(*args)


class SimilarityResponsePayload(Payload):

    msg_id = 2
    format_list = ['H', 'varlenHx20', 'raw']

    def __init__(self, identifier, preference_list, tb_overlap):
        super(SimilarityResponsePayload, self).__init__()
        self.identifier = identifier % 65536
        self.preference_list = preference_list
        self.tb_overlap = tb_overlap

    def to_pack_list(self):
        encoded_tb_overlap = [pack(">20sI", *tb) for tb in self.tb_overlap]
        data = [('H', self.identifier),
                ('varlenHx20', b"".join(self.preference_list)),
                ('raw', b"".join(encoded_tb_overlap))]

        return data

    @classmethod
    def from_unpack_list(cls, identifier, preference_list, tb_overlap):
        args = [identifier,
                [preference_list[i:i + 20] for i in range(0, len(preference_list), 20)],
                [(tb_overlap[i:i + 20], unpack(">I", tb_overlap[i + 20:i + 24])[0])
                 for i in range(0, len(tb_overlap), 24)]]

        return SimilarityResponsePayload(*args)


class PingPayload(Payload):

    msg_id = 3
    format_list = ['H']

    def __init__(self, identifier):
        super(PingPayload, self).__init__()
        self.identifier = identifier % 65536

    def to_pack_list(self):
        data = [('H', self.identifier), ]

        return data

    @classmethod
    def from_unpack_list(cls, identifier):
        return PingPayload(identifier)


class PongPayload(PingPayload):
    msg_id = 4


class DiscoveryIntroductionRequestPayload(IntroductionRequestPayload):

    format_list = ['c20s', '4SH', '4SH', '4SH', 'bits', 'H', 'raw']

    def __init__(self, introduce_to, destination_address, source_lan_address, source_wan_address, advice,
                 connection_type, identifier, extra_bytes):
        super(DiscoveryIntroductionRequestPayload, self).__init__(destination_address, source_lan_address,
                                                                  source_wan_address, advice, connection_type,
                                                                  identifier, extra_bytes)
        self.introduce_to = introduce_to

    def to_pack_list(self):
        data = super(DiscoveryIntroductionRequestPayload, self).to_pack_list()
        data.insert(0, ('c20s', b'Y', self.introduce_to))
        return data

    @classmethod
    def from_unpack_list(cls, introduce_to, destination_address, source_lan_address, source_wan_address,
                         connection_type_0, connection_type_1, dflag0, dflag1, dflag2, tunnel, _, advice,
                         identifier, extra_bytes):
        args = [introduce_to[1:],
                (inet_ntoa(destination_address[0]), destination_address[1]),
                (inet_ntoa(source_lan_address[0]), source_lan_address[1]),
                (inet_ntoa(source_wan_address[0]), source_wan_address[1]),
                [True, False][advice],
                decode_connection_type(connection_type_0, connection_type_1),
                identifier,
                extra_bytes]

        return DiscoveryIntroductionRequestPayload(*args)



from .base import TestBase
from .mocking.endpoint import MockEndpointListener
from .mocking.ipv8 import MockIPv8
from ..community import Community


class OldCommunity(Community):
    community_id = b'\x00' * 20

    def create_introduction_request(self, socket_address, extra_bytes=b'', new_style=False):
        return super().create_introduction_request(socket_address)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier,
                                     introduction=None, extra_bytes=b'', prefix=None, new_style=False):
        return super().create_introduction_response(lan_socket_address, socket_address, identifier, introduction)

    def create_puncture(self, lan_walker, wan_walker, identifier, new_style=False):
        return super().create_puncture(lan_walker, wan_walker, identifier)

    def create_puncture_request(self, lan_walker, wan_walker, identifier, prefix=None, new_style=False):
        return super().create_puncture_request(lan_walker, wan_walker, identifier)


class NewCommunity(Community):
    community_id = b'\x00' * 20


class TestCommunityCompatibility(TestBase):
    """
    Tests for interoperability between old-style and new-style IPv8 Communities.
    """

    def setUp(self):
        super().setUp()
        self.production_overlay_classes.append(OldCommunity)
        self.production_overlay_classes.append(NewCommunity)
        self.make_nodes()

    def make_nodes(self):
        self.nodes = [MockIPv8(u"low", NewCommunity), MockIPv8(u"low", NewCommunity), MockIPv8(u"low", OldCommunity)]
        self.endpoint_listeners = [MockEndpointListener(self.endpoint(i)) for i in range(len(self.nodes))]
        self.new_peer1 = 0
        self.new_peer2 = 1
        self.old_peer = 2

    def endpoint_listener(self, i):
        return self.endpoint_listeners[i]

    def received_message_ids(self, i):
        return [packet[1][22] for packet in self.endpoint_listener(i).received_packets]

    async def walk_from_to(self, from_i, to_i):
        self.overlay(from_i).walk_to(self.address(to_i))
        await self.deliver_messages()

    async def test_introduce_old(self):
        """
        Check that no new-style messages are going to the old-style peer.
        """
        await self.walk_from_to(self.new_peer1, self.new_peer2)
        await self.walk_from_to(self.old_peer, self.new_peer1)

        received_messages_old_peer = self.received_message_ids(self.old_peer)

        self.assertNotIn(231, received_messages_old_peer)
        self.assertNotIn(232, received_messages_old_peer)
        self.assertNotIn(233, received_messages_old_peer)
        self.assertNotIn(234, received_messages_old_peer)



"""
Implementation of proofs for checking commitment equality and if a commitment is a square ("Efficient Proofs that a
Committed NumberLies in an Interval" by F. Boudot).

Modified for use with range proofs ("An efficient range proof scheme." by K. Peng and F. Bao).
"""
from binascii import hexlify
from math import ceil, log
from os import urandom
from struct import pack, unpack

from ..primitives.attestation import sha256_as_int
from ..primitives.structs import ipack, iunpack
from ..primitives.value import FP2Value


def secure_randint(nmin, nmax):
    normalized_range = nmax - nmin
    n = int(ceil(log(normalized_range, 2) / 8.0))
    rbytes_int = int(hexlify(urandom(n)), 16)
    return nmin + (rbytes_int % normalized_range)


def _sipack(*n):
    if len(n) > 8:
        raise RuntimeError("More than 8 values specified to _sipack")
    sign_byte = 0
    packed = b''
    for i in n:
        sign_byte = sign_byte << 1
        sign_byte |= 1 if i < 0 else 0
        packed = ipack(-i if i < 0 else i) + packed
    return pack(">B", sign_byte) + packed


def _siunpack(buf, amount):
    rem = buf[1:]
    nums = []
    sign_byte, = unpack(">B", buf[0:1])
    while rem and len(nums) < amount:
        unpacked, rem = iunpack(rem)
        negative = sign_byte & 0x01
        sign_byte = sign_byte >> 1
        nums.append(-unpacked if negative else unpacked)
    return reversed(nums), rem


class EL(object):

    def __init__(self, c, D, D1, D2):
        self.c = c
        self.D = D
        self.D1 = D1
        self.D2 = D2

    @classmethod
    def create(cls, x, r1, r2, g1, h1, g2, h2, b, bitspace, t=80, l=40):  # pylint: disable=R0913,R0914
        maxrange_w = 2 ^ (l + t) * b - 1
        maxrange_n = 2 ^ (l + t + bitspace) * g1.mod - 1
        w = secure_randint(1, maxrange_w)
        n1 = secure_randint(1, maxrange_n)
        n2 = secure_randint(1, maxrange_n)
        W1 = g1.intpow(w) * h1.intpow(n1)
        W2 = g2.intpow(w) * h2.intpow(n2)
        cW1 = (W1.wp_nominator() * W1.wp_denom_inverse()).normalize()
        cW2 = (W2.wp_nominator() * W2.wp_denom_inverse()).normalize()

        c = sha256_as_int(str(cW1.a).encode('utf-8') + str(cW1.b).encode('utf-8')
                          + str(cW2.a).encode('utf-8') + str(cW2.b).encode('utf-8'))
        D = w + c * x
        D1 = n1 + c * r1
        D2 = n2 + c * r2
        return cls(c, D, D1, D2)

    def check(self, g1, h1, g2, h2, y1, y2):
        cW1 = g1.intpow(self.D) * h1.intpow(self.D1) * y1.intpow(-self.c)
        cW2 = g2.intpow(self.D) * h2.intpow(self.D2) * y2.intpow(-self.c)
        cW1 = (cW1.wp_nominator() * cW1.wp_denom_inverse()).normalize()
        cW2 = (cW2.wp_nominator() * cW2.wp_denom_inverse()).normalize()

        return self.c == sha256_as_int(str(cW1.a).encode('utf-8') + str(cW1.b).encode('utf-8')
                                       + str(cW2.a).encode('utf-8') + str(cW2.b).encode('utf-8'))

    def serialize(self):
        return _sipack(self.c, self.D, self.D1, self.D2)

    @classmethod
    def unserialize(cls, s):
        unpacked, rem = _siunpack(s, 4)
        return cls(*unpacked), rem

    def __eq__(self, other):
        if not isinstance(other, EL):
            return False
        return (self.c == other.c) and (self.D == other.D) and (self.D1 == other.D1) and (self.D2 == other.D2)

    def __hash__(self):
        return 6976

    def __str__(self):
        return 'EL<%d,%d,%d,%d>' % (self.c, self.D, self.D1, self.D2)


class SQR(object):

    def __init__(self, F, el):
        self.F = F
        self.el = el

    @classmethod
    def create(cls, x, r1, g, h, b, bitspace):
        r2 = secure_randint(-2 ^ bitspace * g.mod + 1, 2 ^ bitspace * g.mod - 1)
        F = g.intpow(x) * h.intpow(r2)
        r3 = r1 - r2 * x
        return cls(F, EL.create(x, r2, r3, g, h, F, h, b, bitspace))

    def check(self, g, h, y):
        return self.el.check(g, h, self.F, h, self.F, y)

    def serialize(self):
        min_f = self.F.wp_compress()
        return ipack(min_f.mod) + ipack(min_f.a) + ipack(min_f.b) + self.el.serialize()

    @classmethod
    def unserialize(cls, s):
        rem = s
        mod, rem = iunpack(rem)
        Fa, rem = iunpack(rem)
        Fb, rem = iunpack(rem)
        el, rem = EL.unserialize(rem)
        return cls(FP2Value(mod, Fa, Fb), el), rem

    def __eq__(self, other):
        if not isinstance(other, SQR):
            return False
        return (self.F == other.F) and (self.el == other.el)

    def __hash__(self):
        return 838182

    def __str__(self):
        return 'SQR<%s,%s>' % (str(self.F), str(self.el))



import json
from asyncio import sleep

from ...base import MockIPv8, TestBase
from ....attestation.identity.community import IdentityCommunity
from ....attestation.identity.manager import IdentityManager


class TestIdentityCommunity(TestBase):

    FAKE_HASH = b'a' * 32

    def setUp(self):
        super(TestIdentityCommunity, self).setUp()
        self.initialize(IdentityCommunity, 2)

    def create_node(self):
        identity_manager = IdentityManager(u":memory:")
        return MockIPv8(u"curve25519", IdentityCommunity, identity_manager=identity_manager)

    async def test_advertise(self):
        """
        Check if a node can construct an advertisement for his attested attribute.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(1, len(pseudonym.get_credentials()))
            self.assertEqual(1, len(pseudonym.get_credentials()[0].attestations))

    async def test_advertise_twice(self):
        """
        Check if a node can construct two attributes.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        self.overlay(1).add_known_hash(self.FAKE_HASH[:-1] + b'b', "attribute", self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(2, len(pseudonym.get_credentials()))
            self.assertEqual(1, len(pseudonym.get_credentials()[0].attestations))
            self.assertEqual(1, len(pseudonym.get_credentials()[1].attestations))

    async def test_advertise_big(self):
        """
        Check if a node can construct an advertisement for his attested attribute, having more than 32 attributes.
        """
        await self.introduce_nodes()

        for i in range(39):
            self.overlay(0).self_advertise(self.FAKE_HASH[:-1] + bytes([i]), "attribute" + str(i))
        self.overlay(1).add_known_hash(self.FAKE_HASH[:-1] + bytes([39]), "attribute" + str(39), self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH[:-1] + bytes([39]),
                                                          "attribute" + str(39))

        await self.deliver_messages()
        await sleep(.5)

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(40, len(pseudonym.tree.elements))
            self.assertEqual(40 if node_nr == 0 else 1, len(pseudonym.get_credentials()))

    async def test_advertise_metadata(self):
        """
        Check if a node can construct an advertisement for his attested attribute with metadata.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(0), {"a": "b"})
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute", "id_metadata",
                                                          {"a": "b"})

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(1, len(pseudonym.get_credentials()))
            self.assertEqual(1, len(pseudonym.get_credentials()[0].attestations))
            self.assertIn("a", json.loads(pseudonym.get_credentials()[0].metadata.serialized_json_dict))
            self.assertEqual("b", json.loads(pseudonym.get_credentials()[0].metadata.serialized_json_dict)["a"])

    async def test_advertise_metadata_reject(self):
        """
        Check if a node cannot construct an advertisement for his attested attribute with wrong metadata.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(0), {"c": "d"})
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute", "id_metadata",
                                                          {"a": "b"})

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(1, len(pseudonym.get_credentials()))
            self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))

    async def test_advertise_reject_hash(self):
        """
        Check if unknown hashes are not signed.
        """
        await self.introduce_nodes()

        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        pseudonym = self.overlay(0).identity_manager.get_pseudonym(self.public_key(0))
        self.assertEqual(1, len(pseudonym.get_credentials()))
        self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))

        pseudonym = self.overlay(1).identity_manager.get_pseudonym(self.public_key(0))
        self.assertEqual(0, len(pseudonym.get_credentials()))

    async def test_advertise_reject_public_key(self):
        """
        Check if we don't sign correct hashes for the wrong peer.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(1))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        pseudonym = self.overlay(0).identity_manager.get_pseudonym(self.public_key(0))
        self.assertEqual(1, len(pseudonym.get_credentials()))
        self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))

        pseudonym = self.overlay(1).identity_manager.get_pseudonym(self.public_key(0))
        self.assertEqual(0, len(pseudonym.get_credentials()))

    async def test_advertise_reject_old(self):
        """
        Check if we don't sign old attestations.
        """
        await self.introduce_nodes()

        self.overlay(1).known_attestation_hashes[self.FAKE_HASH] = ("attribute", 0, self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attribute")

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(1, len(pseudonym.get_credentials()))
            self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))

    async def test_advertise_reject_wrong_name(self):
        """
        Check if we don't sign attestations with incorrect metadata.
        """
        await self.introduce_nodes()

        self.overlay(1).add_known_hash(self.FAKE_HASH, "attribute", self.key_bin(0))
        self.overlay(0).request_attestation_advertisement(self.peer(1), self.FAKE_HASH, "attr1bute")

        await self.deliver_messages()

        for node_nr in [0, 1]:
            pseudonym = self.overlay(node_nr).identity_manager.get_pseudonym(self.public_key(0))

            self.assertEqual(1, len(pseudonym.get_credentials()))
            self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))



from ...base import TestBase
from ....messaging.interfaces.network_stats import NetworkStat


class TestNetworkStat(TestBase):

    async def test_initialize(self):
        """
        Check if the starting values of NetworkStat are sane.
        """
        stats = NetworkStat(1447)

        self.assertEqual(1447, stats.identifier)
        self.assertEqual(0, stats.num_up)
        self.assertEqual(0, stats.num_down)
        self.assertEqual(0, stats.bytes_up)
        self.assertEqual(0, stats.bytes_down)
        self.assertEqual(0, stats.first_measured_up)
        self.assertEqual(0, stats.first_measured_down)
        self.assertEqual(0, stats.last_measured_up)
        self.assertEqual(0, stats.last_measured_down)

    async def test_add_sent_stat(self):
        """
        Check if an added sent statistic is correctly registered.
        """
        stats = NetworkStat(0)

        stats.add_sent_stat(1492, 1776)

        self.assertEqual(1, stats.num_up)
        self.assertEqual(1776, stats.bytes_up)
        self.assertEqual(1492, stats.first_measured_up)
        self.assertEqual(1492, stats.last_measured_up)

    async def test_add_sent_stat_first_up(self):
        """
        Check if the first_measured_up is correctly registered on send.
        """
        stats = NetworkStat(0)

        stats.add_sent_stat(1492, 1776)
        stats.add_sent_stat(1619, 1776)

        self.assertEqual(1492, stats.first_measured_up)

    async def test_add_received_stat(self):
        """
        Check if an added received statistic is correctly registered.
        """
        stats = NetworkStat(0)

        stats.add_received_stat(1492, 1776)

        self.assertEqual(1, stats.num_down)
        self.assertEqual(1776, stats.bytes_down)
        self.assertEqual(1492, stats.first_measured_down)
        self.assertEqual(1492, stats.last_measured_down)

    async def test_add_received_stat_first_down(self):
        """
        Check if the first_measured_up is correctly registered on receive.
        """
        stats = NetworkStat(0)

        stats.add_received_stat(1492, 1776)
        stats.add_received_stat(1619, 1776)

        self.assertEqual(1492, stats.first_measured_down)

    async def test_to_dict(self):
        """
        Check if the dictionary form of the NetworkStat is correct.
        """
        stats = NetworkStat(1849)
        stats.add_sent_stat(1492, 1776)
        stats.add_received_stat(1619, 1787)
        stats.add_sent_stat(1783, 1794)
        stats.add_received_stat(1789, 1798)

        self.assertDictEqual({
            "identifier": 1849,
            "num_up": 2,
            "num_down": 2,
            "bytes_up": 1776 + 1794,
            "bytes_down": 1787 + 1798,
            "first_measured_up": 1492,
            "first_measured_down": 1619,
            "last_measured_up": 1783,
            "last_measured_down": 1789
        }, stats.to_dict())

    async def test_to_str(self):
        """
        Check if a NetworkStat is correctly formatted as a str.
        """
        stats = NetworkStat(1849)
        stats.add_sent_stat(1492, 1776)
        stats.add_received_stat(1619, 1787)
        stats.add_sent_stat(1783, 1794)
        stats.add_received_stat(1789, 1798)

        self.assertEqual("NetworkStat{num_up:2, num_down:2, bytes_up:3570, bytes_down:3585, ...}", str(stats))
